#!/bin/bash
#vim: set expandtab cindent fdm=marker ts=2 sw=2:

# vim       :ShellCheck
# terminal  shellcheck common_library.sh
LANG=C
export PS4='$0.$LINENO+ '

export HWLOC_ALLOW=all

RPM_INSTALL="yum --skip-broken -y install"
[[ -f /etc/fedora-release ]] && RELEASE=$(grep -Eho "[0-9]+" /etc/fedora-release) && (( RELEASE > 21 )) && RPM_INSTALL="dnf --setopt=strict=0 -y install" && RPM_REMOVE="dnf -y remove"
export RPM_INSTALL
export RPM_REMOVE="yum -y remove"

#{{{ function install_regular_then_epel
function install_regular_then_epel() {
  if [[ -z "$RPM_INSTALL" ]]; then echo "${RPM_INSTALL} is unset, function install_regular_then_epel cannot work"; fi
  for RPM_PACKAGE in "$@"
  do
    eval "$RPM_INSTALL" "$RPM_PACKAGE"

    if ! rpm -q "$RPM_PACKAGE" >/dev/null 2>&1; then
      eval "$RPM_INSTALL" --enablerepo epel "$RPM_PACKAGE"
    fi

    if ! rpm -q "$RPM_PACKAGE" >/dev/null 2>&1; then
      echo "ERROR: Installation of $RPM_PACKAGE has failed! Tried both default repositories and epel, still failing"
    fi

  done
}
#}}}

#{{{function howManyElements
howManyElements()
{
  echo $#
}
#}}}

#{{{function abs
abs()
{
   [ "$1" -lt 0 ] && echo $((-$1)) || echo "$1"
}
#}}}function abs

#{{{function pathadd
# See https://superuser.com/questions/39751/add-directory-to-path-if-its-not-already-there/39995
pathadd()
{
    if [ -d "$1" ] && [[ ":$PATH:" != *":$1:"* ]]; then
        PATH="${PATH:+"$PATH:"}$1"
    fi
}
#}}}function pathadd

# {{{ tarxz
tarxz()
{
  if [[ $# == 0 ]]; then
    echo "Usage tarxz directory/file" 1>&2;
    echo "Or    tarxz output.tar.xz directory/file directory/file ..." 1>&2;
    return 1;
  else
    if [[ $# == 1 ]]; then
      OUTPUT="$(dirname "$1")/$(basename "$1").tar.xz"
      OPTION="--directory=$(dirname "$1")"
      DATA=$(basename "$1")
    else
      OUTPUT=$1
      OPTION=""
      DATA=""
      shift
    fi
  fi

  if [[ -e "${OUTPUT}" ]]; then
    echo "File \"${OUTPUT}\" exists already" 1>&2
    return 1
  fi

  if [ "${DATA}xx" != "xx" ]; then
    XZ_OPT=-T0 tar -Jcf "$OUTPUT" "$OPTION" "$DATA"
  else
    XZ_OPT=-T0 tar -Jcf "$OUTPUT" "$@"
  fi
}
# }}} tarxz

#{{{ saveFile
# Use cat to save /sys files
saveFile()
{
   local DEST="$1"
   local FILE="$2"
   DIR=$(dirname "${FILE}")
   mkdir -p "${DEST}/${DIR}" 2>/dev/null
   cat "${FILE}" > "${DEST}/${FILE}" 2>/dev/null
}

#}}}

#{{{ saveLink
saveLink()
{
   local DEST="$1"
   local LINK="$2"
   DIR=$(dirname "${FILE}")
   mkdir -p "${DEST}/${DIR}" 2>/dev/null
   cp -P "${LINK}" "${DEST}/${LINK}"
}

#}}}

#{{{ saveDir
# Copy all files from the given directory
saveDir()
{
   local DEST="$1"
   local SOURCE="$2"
# gather all directories, including empty ones
   find "${SOURCE}" -type d 2>/dev/null | while read -r DIR ; do
      mkdir -p "${DEST}/${DIR}" 2>/dev/null
   done
# gather all files now
   find "${SOURCE}" -type f 2>/dev/null | while read -r FILE ; do
      saveFile "${DEST}" "${FILE}"
   done
# gather symlinks
   find "${SOURCE}" -type l 2>/dev/null | while read -r LINK ; do
      saveLink "${DEST}" "${LINK}"
   done
}

#}}}

#{{{ function my_echo
my_echo() {
  echo
  echo "==================================================================================================="
  for S in "$@"; do
    echo "$(date +"%T") $S"
  done
  echo "==================================================================================================="
  echo
}
#}}}

#{{{function retry
function retry {
  local n=1
  local max=4
  local delay=600
  while true; do
    "$@"
    status=$?
    if [[ $status == 0 ]]; then
      break
    else
      if [[ $n -lt $max ]]; then
        ((n++))
        echo "Command \"$*\" failed. Sleeping for $delay seconds."
        sleep $delay
        delay=$((delay*2))
        echo "Attempt $n/$max."
      else
        echo "The command \"$*\" has failed after $n attempts. Last exit status was $status."
        return $status
      fi
    fi
  done
}
#}}}

#{{{ ceiling_divide
function ceiling_divide()
{
  local num=$1
  local div=$2
  echo $(( (num + div -1) / div ))
}
#}}} ceiling_divide

#{{{ returnHTString
function returnHTString()
{
  HT=$(lscpu |grep "Thread(s)"|awk '{print $NF}')

  if [[ -z "$HT" ]];  then
    echo >&2 "function returnHTString - cannot determine the number of Thread(s) per core"
    echo "htUNKNOWN"
  else
    if (( HT >= 2 )); then
      echo "htON"
    else
      echo "htOFF"
    fi
  fi
}
# }}} returnHTString

# {{{ storeXMLs
function storeXMLs()
{
  my_echo "storeXMLs"
  if [[ -n "$BEAKER_HUB_URL" ]] && [[ -n "$RSTRNT_RECIPEID" ]]; then
    if ! wget "$BEAKER_HUB_URL/recipes/${RSTRNT_RECIPEID}.xml" -O "${CONFIG_DIR}/job.xml"; then
      echo
        echo "wget $BEAKER_HUB_URL/recipes/${RSTRNT_RECIPEID}.xml -O ${CONFIG_DIR}/job.xml has failed"
      fi
  fi
}
# }}} storeXMLs

# {{{ function compareVersions ver1 ver2
# USAGE
#    compareVersions $1 $2
#    case $? in
#        0) op='=';;
#        11) op='>';;
#        12) op='<';;
#    esac

function compareVersions()
{
  if command -v rpmdev-vercmp >/dev/null 2>&1; then
    rpmdev-vercmp "$1" "$2"
    return $?
  else
    if [[ "$1" == "$2" ]]; then
      return 0
    fi

    local version=$1 check=$2
    local winner
    winner=$(echo -e "$version\n$check" | sed '/^$/d' | sort -t. -nr | head -1)

    if [[ "$winner" == "$version" ]]; then
      return 11
    else
      return 12
    fi
  fi
}
#}}} compareVersions

#{{{ copy config files
function copyConfigFiles()
{
  my_echo "hwloc-gather-topology"
  hwloc-gather-topology "${CONFIG_DIR}"/topology-archive
  sysctl -a > "${CONFIG_DIR}"/sysctl.txt

  my_echo "Copying /sys directories"
  #Copy /sys directories
  cat << EOF | while read -r DIRS ; do saveDir "${CONFIG_DIR}" "${DIRS}" ; done
/sys/devices/system/cpu/
/sys/devices/system/node/
/sys/bus/cpu/devices/
/sys/bus/node/devices/
/sys/class/dmi/id/
/sys/kernel/mm/hugepages/
/sys/kernel/mm/transparent_hugepage/
/proc/device-tree/cpus/
EOF

  #Copy some /proc files
  cat << EOF | while read -r FILES ; do saveFile "${CONFIG_DIR}" "${FILES}" ; done
/proc/cpuinfo
/proc/meminfo
/proc/mounts
/proc/stat
/proc/slabinfo
/proc/self/cpuset
/proc/cmdline
EOF

  uname -a >> "${CONFIG_DIR}"/uname-a.txt 2>&1
  cat /proc/meminfo >> "${CONFIG_DIR}"/meminfo.txt

  if command -v tuned-adm >/dev/null 2>&1; then
    tuned-adm active | head -1 | awk -F' ' '{print $4}' >> "${CONFIG_DIR}"/tuned-active-profile.txt
  else
    echo "NOPROFILE" >> "${CONFIG_DIR}"/tuned-active-profile.txt
  fi

  (command -v getenforce >/dev/null 2>&1 && getenforce || echo "UNKNOWN" )                   &>  "${CONFIG_DIR}"/selinux.txt
  (command -v sestatus >/dev/null 2>&1   && sestatus   || echo "sestatus is not installed" ) &>  "${CONFIG_DIR}"/selinux_full_log.txt
}
#}}} copy config files

#{{{ reportVulnerabilities - Report known vulnerabilities and their status
function reportVulnerabilities() {
# shellcheck disable=SC2012
if [[ -d "/sys/devices/system/cpu/vulnerabilities" ]] && [[ "$(ls -A "/sys/devices/system/cpu/vulnerabilities")" ]] ; then
  pushd "/sys/devices/system/cpu/vulnerabilities" > /dev/null || exit
# shellcheck disable=SC2046
  tail -v -n +1 $(ls | sort)
  popd >/dev/null || exit
fi
}
#}}}

#{{{function getSocketsNo
function getSocketsNo()
{
  local noOfSockets
  noOfSockets=$(hwloc-calc -N Package machine:0)
  if [ -z "${noOfSockets}" ]; then
    noOfSockets=$(hwloc-calc -N Socket machine:0)
  fi
  echo "$noOfSockets"
}
#}}}

#{{{ function getNumasNo
function getNumasNo()
{
  numastat | head -1 | wc -w
}
#}}} function getNumasNo

# {{{ function getNoCores()
function getNoCores() {
  hwloc-calc -N core machine:0
}
# }}} function getNoCores()

# {{{ function getNoCpus()
function getNoCpus() {
  hwloc-calc -N PU machine:0
}
# }}} function getNoCpus()

# {{{ function getNoThreads
function getNoThreads() {
  lscpu |grep "Thread(s)"|awk '{print $NF}'
}
# }}} function getNoThreads()

#{{{ getNumaNodeFromCPU
# parameter cpuNumber
function getNumaNodeFromCPU()
{
  local CPUNUMBER NODEDIRS ANODEDIR ANODE
  CPUNUMBER=$1
  NODEDIRS=$(echo /sys/devices/system/node/node*)

  for ANODEDIR in $NODEDIRS; do
    ANODE=$(basename "$ANODEDIR" | sed 's/node//')

    if [ -e "$ANODEDIR/cpu${CPUNUMBER}" ]; then
      echo "$ANODE"
      return
    fi
  done
}
#}}} getNumaNodeFromCPU

#{{{ hexMaskToCpuNo - sets global variables SOCKET CORE and PU
# parameter hexNumber
function hexMaskToCpuNo()
{
  SOCKET=$(hwloc-calc -p "$1" --hierarchical Package|awk -F":" '{print $2}')

  if [ -z "$SOCKET" ]; then
    SOCKET=$(hwloc-calc -p "$1" --hierarchical socket|awk -F":" '{print $2}')
  fi

  (( SOCKET==4294967295 )) || (( SOCKET==-1 )) && SOCKET=0

  CORE=$(hwloc-calc -p "$1" --hierarchical core|awk -F":" '{print $2}')
  PU=$(hwloc-calc -p "$1" --hierarchical PU|awk -F":" '{print $2}')

  export SOCKET
  export CORE
  export PU
}
#}}} hexMaskToCpuNo

#{{{function ord
ord() {
  LC_CTYPE=C printf '%d' "'$1"
}
#}}}

#{{{ function max using sort
# Use -g for --general-numeric-sort
#max -g 1.5 9 -8 10.3 1e9

function max() {
  printf "%s\n" "${@:2}" | sort "$1" | tail -n1
}
#}}}

#{{{ function min using sort
# Use -g for --general-numeric-sort
#min -g 1.5 9 -8 10.3 1e9

function min() {
  printf "%s\n" "${@:2}" | sort --reverse "$1" | tail -n1
}
#}}}

#{{{ function sort_uniq()
function sort_uniq() {
  printf "%s\n" "${@}" | sort -gu
}
#}}}

#{{{ function highestPowerof2
function highestPowerof2() {
  local res i n

  n=$1
  res=0

  for (( i=1; i<=n; i*=2)); do
    if (( i>n )); then
      break
    fi
    res=$i
  done
  echo "$res"
}
#}}}

#{{{ function add
#See https://unix.stackexchange.com/questions/360324/addition-of-two-floating-point-numbers-using-shell-script
function add() {
  if [[ $# -eq 0 ]]; then
    echo 0
  else
    n="$*"; bc <<< "${n// /+}"
  fi
}
#}}}

#{{{ function convertsecs()
function convertsecs() {
 ((d=${1}/86400))
 ((h=(${1}%86400)/3600))
 ((m=(${1}%3600)/60))
 ((s=${1}%60))
 ((d != 0)) && { printf "%02d days %02dh:%02dm:%02ds\n" "$d" "$h" "$m" "$s"; return; }
 ((h != 0)) && { printf "%02dh:%02dm:%02ds\n" "$h" "$m" "$s"; return; }
 ((m != 0)) && { printf "%02dm:%02ds\n" "$m" "$s"; return; }
 printf "%02ds\n" "$s"
}
#}}}

#{{{ function check_if_installed
function check_if_installed() {
  local tool
  local return_status
  return_status=0
  for tool in "$@"; do
    command -v "$tool" >/dev/null 2>&1 || { echo >&2 "$tool is required, but it's not installed."; return_status=1; }
  done
  return "$return_status"
}
#}}}

# {{{ function mpstat_add_NUMA_info "$input_dir" "$lscpu_file" "$output_dir"
# The function body can be any compound command, not just a group command ( {} ). Use a sub-shell:
# See https://stackoverflow.com/questions/12179633/make-shopt-change-local-to-function

function mpstat_add_NUMA_info() (
  local input_dir
  local lscpu_file
  local output_dir
  local binary

  [ "$#" -ne 3 ] && { echo "function mpstat_add_NUMA_info - expecting 3 arguments, got $#."; printf '"%s"\n' "${@}"; echo "Returning from function."; return 1; }
  input_dir="$1"
  lscpu_file="$2"
  output_dir="$3"

  binary="${TESTDIR:-.}/mpstat2node.py"
  [[ ! -e "$binary" ]] && { echo >&2 "ERROR - function mpstat_add_NUMA_info: $binary does not exist."; return 1; }
  [[ ! -x "$binary" ]] && { echo >&2 "ERROR - function mpstat_add_NUMA_info: $binary is not executable."; return 1; }

  if [[ -n "$output_dir" ]]; then
    my_echo "mpstat_add_NUMA_info"
# This change is local - we are in subshell
    shopt -s nullglob
    for file in "$input_dir"/*mpstat; do
      temp_file=$(mktemp)
      if "$binary" --lscpu "$lscpu_file" <"$file" >"$temp_file"
      then
        mv "$temp_file" "$output_dir/$(basename "$file").NUMA"
      else
        echo "$binary has failed for file \"${file}\""
        rm "$temp_file"
      fi
    done
  fi
)
# }}}

# {{{ function ps_add_NUMA_info "$input_dir" "$lscpu_file" "$output_dir"
# The function body can be any compound command, not just a group command ( {} ). Use a sub-shell:
# See https://stackoverflow.com/questions/12179633/make-shopt-change-local-to-function

function ps_add_NUMA_info() (
  local input_dir
  local lscpu_file
  local output_dir
  local out_file
  local ps2node_bin
  local ps2hist_bin

  [ "$#" -ne 3 ] && { echo "function ps_add_NUMA_info - expecting 3 arguments, got $#."; printf '"%s"\n' "${@}"; echo "Returning from function."; return 1; }
  input_dir="$1"
  lscpu_file="$2"
  output_dir="$3"

  ps2node_bin="${TESTDIR:-.}/ps2node.py"
  [[ ! -e "$ps2node_bin" ]] && { echo >&2 "ERROR - function ps_add_NUMA_info: $ps2node_bin does not exist."; return 1; }
  [[ ! -x "$ps2node_bin" ]] && { echo >&2 "ERROR - function ps_add_NUMA_info: $ps2node_bin is not executable."; return 1; }

  ps2hist_bin="${TESTDIR:-.}/ps2hist.py"
  [[ ! -e "$ps2hist_bin" ]] && { echo >&2 "ERROR - function ps_add_NUMA_info: $ps2hist_bin does not exist."; return 1; }
  [[ ! -x "$ps2hist_bin" ]] && { echo >&2 "ERROR - function ps_add_NUMA_info: $ps2hist_bin is not executable."; return 1; }

  if [[ -n "$output_dir" ]]; then
    my_echo "ps_add_NUMA_info"
#This change is local - we are in subshell
    shopt -s nullglob
    for file in "$input_dir"/*thread_log; do
      temp_file=$(mktemp)
      if "$ps2node_bin" --lscpu "$lscpu_file" <"${file}" >"${temp_file}"
      then
        out_file="$output_dir/$(basename "$file")"
        mv "$temp_file" "$out_file"
        "$ps2hist_bin" <"$out_file" >"$out_file".numa.hist
      else
        echo "$ps2node_bin has failed for file \"${file}\""
        rm "$temp_file"
      fi
    done
  fi
)
# }}}

# {{{ function check_for_numatop_and_numastat - sets global variables RUN_NUMATOP and RUN_NUMASTAT
# check if it makes sense to run numatop and numastat
function check_for_numatop_and_numastat() {
  local numatop_pid
  local numa_nodes

  export RUN_NUMATOP=0
  export RUN_NUMASTAT=0
  numa_nodes=$(getNumasNo)

# Must be NUMA server, numatop has to be installed and support the CPU model
  if (( numa_nodes > 1 )) ; then
    export RUN_NUMASTAT=1

    if ! check_if_installed numatop ;then
      echo >&2 "Attempting automatic installation via \"${RPM_INSTALL} numatop\""
      ${RPM_INSTALL} numatop

      if ! command -v numatop >/dev/null 2>&1 ;then
        echo >&2 "\"${RPM_INSTALL} numatop\" has failed, attempting automatic installation from https://github.com/01org/numatop.git"

        pushd "$(mktemp -d -p ./ GIT_NUMATOP_"$(date +%Y-%b-%d_%Hh%Mm%Ss)".XXX)" || exit
        git clone https://github.com/01org/numatop.git
        SET_SETTING=$(shopt -po errexit)
        set +e
        [ -d numatop ] && { cd numatop && ./autogen.sh && make && make install ; }
        eval "$SET_SETTING"
        popd || exit
      fi
      command -v numatop >/dev/null 2>&1 || { echo >&2 "Automatic installation of numatop has failed."; }
    fi

    if command -v numatop >/dev/null 2>&1 ; then
      echo "N" | numatop -d /dev/null >/dev/null 2>&1 &
      numatop_pid=$!
      sleep 1
      if ps -p "$numatop_pid" &>/dev/null ; then
        kill "$numatop_pid"
        export RUN_NUMATOP=1
      fi
    fi
  fi
}
#}}}

#{{{ function get_distro_name - sets global variable DISTRO_NAME
function get_distro_name {
  local distro_from_motd

  DISTRO_NAME=${RSTRNT_OSDISTRO}

  if [[ -z "${DISTRO_NAME}" ]]; then
    if distro_from_motd=$(grep -E -m 1 -o "DISTRO=[^ ]+" /etc/motd) ; then
      DISTRO_NAME=$(awk -F = '{print $2}' <<<"$distro_from_motd")
    fi
  fi
  if [[ -z "$DISTRO_NAME" ]]; then
    if [[ -r /etc/redhat-release ]]; then
      DISTRO_NAME=$(</etc/redhat-release perl -pl -e's/\(.*\)/ /'| perl -pl -e'tr/ /-/'|perl -pl -e's/--$//')
    else
      DISTRO_NAME="unknown"
    fi
  fi
  export DISTRO_NAME
}
#}}}

#{{{ function get_jobid_name - sets global variable JOBID_NAME
function get_jobid_name {
  local from_motd

  JOBID_NAME=${RSTRNT_JOBID}

  if [[ -z "${JOBID_NAME}" ]]; then
    if from_motd=$(grep -E -m 1 -o "JOBID=[^ ]+" /etc/motd) ; then
      JOBID_NAME=$(awk -F = '{print $2}' <<<"$from_motd")
    fi
  fi
  if [[ -n "$JOBID_NAME" ]]; then
    export JOBID_NAME
  fi
}
#}}}

#{{{ function get_hosttype_and_virttype - sets global variables VIRT_TYPE and HOST_TYPE
function get_hosttype_and_virttype {
  local virt_type_temp

  VIRT_TYPE=""
  if grep -q "^flags.*hypervisor" /proc/cpuinfo ; then
    HOST_TYPE="VIRTUAL"
    if command -v virt-what >/dev/null 2>&1 ; then
      if virt_type_temp=$(virt-what) ; then
        VIRT_TYPE=$virt_type_temp
      fi
    fi
  else
    HOST_TYPE="MACHINE"
  fi
  export HOST_TYPE
  export VIRT_TYPE
}
#}}}

#{{{ function get_uptime
function get_uptime {
  local date_mili
  local uptime
  local date_seconds
  local uptime_only
  local trials
  local results
  local datamash
  local i

  results=()

  if command -v datamash >/dev/null 2>&1; then
    trials=15
    datamash=1
  else
    trials=1
    datamash=0
  fi

# shellcheck disable=SC2034  # Unused variables left for readability
  for i in $(seq "$trials"); do
    date_mili=$(date +%s%3N); uptime=$(cat /proc/uptime)
    date_seconds=$(bc -l <<< "scale=3;${date_mili}/10^3")
    uptime_only=$(awk '{print $1}' <<<"$uptime")
    results+=("$(bc <<< "${date_seconds}-${uptime_only}")")
    sleep 0.1
  done

  if (( datamash == 1 )); then
    printf "%s\n" "${results[@]}" | datamash median 1
  else
    echo "${results[0]}"
  fi
}
#}}}

#{{{ function store_tag
function store_tag {
  local file whiteboard regex

  [ "$#" -ne 2 ] && { echo "function store_tag - expecting 2 arguments, got $#."; printf '"%s"\n' "${@}"; echo "Returning from function."; return 1; }

  whiteboard=$1
  file=$2

  regex='"SpecificTag": "([^[:space:]"]*)"'
  if [[ $whiteboard =~ $regex ]] ; then
      echo "${BASH_REMATCH[1]}" >> "${file}"
  fi
}
#}}}

#{{{ function get_memcpy_method
function get_memcpy_method  {
  local method DIR
  DIR=${CONFIG_DIR:-.}

  # debuginfo-install -y glibc
  # yum -y install systemtap-devel systemtap kernel-headers kernel-devel kernel-debuginfo systemtap-runtime
  cat << EOF >"$DIR/SystemTap.stp"
global saw

function foo()
{
  if ( ! saw[ppfunc()]) {
    printf("Detected %s\n", ppfunc());
    saw[ppfunc()] = 1;
  }
}

probe process("/lib64/libc.so.6").function("__memcpy_avx512_no_vzeroupper")? { foo(); }
probe process("/lib64/libc.so.6").function("__memcpy_avx512_unaligned")? { foo(); }
probe process("/lib64/libc.so.6").function("__memcpy_avx_unaligned")? { foo(); }
probe process("/lib64/libc.so.6").function("__memcpy_avx_unaligned_erms")? { foo(); }
probe process("/lib64/libc.so.6").function("__memcpy_avx_unaligned")? { foo(); }
probe process("/lib64/libc.so.6").function("__memcpy_avx_unaligned_erms")? { foo(); }
probe process("/lib64/libc.so.6").function("__memcpy_erms")? { foo(); }
probe process("/lib64/libc.so.6").function("__memcpy_sse2_unaligned")? { foo(); }
probe process("/lib64/libc.so.6").function("__memcpy_sse2_unaligned_erms")? { foo(); }
probe process("/lib64/libc.so.6").function("__memcpy_ssse3")? { foo(); }
probe process("/lib64/libc.so.6").function("__memcpy_ssse3_back")? { foo(); }
EOF

  if stap -vv "$DIR/SystemTap.stp" -c '/bin/ls -l' -o "${DIR}/SystemTap.out" >"${DIR}/SystemTap.log" 2>&1; then
    if [[ -s "${DIR}/SystemTap.out" ]]; then
      method=$(awk 'match($0, /Detected __(memcpy.*)$/, a) {print a[1]}' < "${DIR}/SystemTap.out")
    else
      echo >&2 "Failed to parse memcpy method from '${DIR}/SystemTap.txt'"
      tail -v -n +1 "${DIR}/SystemTap.txt" >&2
    fi
  else
    echo >&2 "stap -vv memcpy.stp has failed!"
  fi
  echo "${method:-UNKNOWN}"
}
#}}}

#{{{ SaveSystemParameters
#SaveSystemParameters "$BENCHMARK_NAME" "$DISTRO_NAME" "$HOSTNAME" "$HTString" "$RESULTSTIME" "$RESULTTUNEPROFILE" "$HOST_TYPE" "$VIRT_TYPE"
function SaveSystemParameters()
{
  local KERNEL_RPM_NAME
  local file
  local SRC_FULL_NAME
  local UNAME_DATE
  local UNAME_DATE_SECONDS
  local MIN_TIME_DIFFERENCE
  local INDEX
  local INDEX_COUNTER
  local RPM_INFO
  local BUILD_DATE_SECONDS
  local TIME_DIFFERENCE
  local SRC_NAME
  local UMOUNT_DEBUGFS

  SRC_FULL_NAME=$(rpm -q -i kernel-"$(uname -r)" | grep -o "Source RPM.*" | awk 'BEGIN {FS=": "} {print $2}')
  [[ -n "$SRC_FULL_NAME" ]] &&  KERNEL_RPM_NAME=$(rpm -q kernel | grep "$(uname -r)")

  if [[ -z "$SRC_FULL_NAME" ]]; then
# CODE above has failed probably because of the uncommon kernel name.
# Try to get Kernel RPM name by finding RPM which containes /boot/vmlinuz-$(uname -r) or /boot/config-$(uname -r) files
    for file in /boot/vmlinuz-$(uname -r) /boot/config-$(uname -r); do
      if [[ -f "$file" ]] ; then
        SRC_FULL_NAME=$(rpm -qif "$file" | grep -o "Source RPM.*" | awk 'BEGIN {FS=": "} {print $2}')
        if [[ -n "$SRC_FULL_NAME" ]]; then
          KERNEL_RPM_NAME=$(rpm -qf "$file")
          echo "INFO - Source RPM was detected to be \"$SRC_FULL_NAME\" based on running rpm -qif \"$file\""
          break
        else
          echo "WARNING - could find to which RPM package the file \"$file\" belongs"
          echo "==================================DEBUG START==========================================="
          echo "rpm -qif \"$file\" output:"
          rpm -qif "$file"
          echo "==================================DEBUG END============================================="
        fi
      else
        echo "WARNING - file \"$file\" does not exist!"
      fi
    done
  fi

  if [[ -z "$SRC_FULL_NAME" ]]; then
# CODE above has failed probably because of the uncommon kernel name.
# As the last resort, let's try to find the SRPM for the running kernel by checking the kernel build time
    UNAME_DATE=$(uname -v | grep -Eoi '(Mon|Tue|Wed|Thu|Fr|Sat|Sun).*')
    UNAME_DATE_SECONDS=$(date -d "${UNAME_DATE}" +%s)

    MIN_TIME_DIFFERENCE=2147483647
    INDEX=0
    INDEX_COUNTER=1
    RPM_INFO=$(rpm -qi kernel)
    while read -r BUILD_DATE; do
      BUILD_DATE_SECONDS=$(date -d "${BUILD_DATE}" +%s)
      #date -d "${BUILD_DATE}"
      TIME_DIFFERENCE=$(abs $((BUILD_DATE_SECONDS-UNAME_DATE_SECONDS)))
      #echo -e "$BUILD_DATE\t$UNAME_DATE\t$TIME_DIFFERENCE"
      #echo -e "$BUILD_DATE_SECONDS\t$UNAME_DATE_SECONDS"
      if (( TIME_DIFFERENCE < MIN_TIME_DIFFERENCE )) ; then
        MIN_TIME_DIFFERENCE=${TIME_DIFFERENCE}
        INDEX=${INDEX_COUNTER}
        #echo -e "$MIN_TIME_DIFFERENCE\t${INDEX}"
      fi
      ((INDEX_COUNTER++))
    done <<< "$(echo "${RPM_INFO}" | grep -i "Build Date" | cut -d: -f2-)"

    if (( INDEX > 0 )) ; then
      SRC_FULL_NAME=$(echo "${RPM_INFO}" | grep -m${INDEX} "^Source RPM" | tail -n1 | awk 'BEGIN {FS=": "} {print $2}')
      KERNEL_RPM_VERSION=$(echo "${RPM_INFO}" | grep -m${INDEX} "^Version" | tail -n1 | awk 'BEGIN {FS=": "} {print $2}')
      KERNEL_RPM_RELEASE=$(echo "${RPM_INFO}" | grep -m${INDEX} "^Release" | tail -n1 | awk 'BEGIN {FS=": "} {print $2}')
      KERNEL_RPM_ARCH=$(echo "${RPM_INFO}" | grep -m${INDEX} "^Architecture" | tail -n1 | awk 'BEGIN {FS=": "} {print $2}')
      KERNEL_RPM_NAME="kernel-${KERNEL_RPM_VERSION}-${KERNEL_RPM_RELEASE}.${KERNEL_RPM_ARCH}"
      echo "INFO - minimum time difference between uname -v and rpm build time found to be ${MIN_TIME_DIFFERENCE} seconds for ${SRC_FULL_NAME}"
    else
      echo "ERROR - Searching for the SRPM for the running kernel by checking the kernel build time - NO MATCH FOUND"
      echo "Output for rpm -qi kernel is"
      echo "=================================================================================="
      echo "${RPM_INFO}"
      echo "=================================================================================="
    fi
  fi

  if [[ -z "$KERNEL_RPM_NAME" ]] ; then
    echo "ERROR - failed to find the kernel RPM name. Exiting"
    echo "Dumping all shell variables"
    env
    exit 1
  else
    echo "INFO - KERNEL_RPM_NAME is \"$KERNEL_RPM_NAME\", running kernel is \"$(uname -r)\""
  fi

  SRC_NAME=$(basename "${SRC_FULL_NAME}" .src.rpm)

  UMOUNT_DEBUGFS=0
  mount | grep -q debugfs
  [[ ${PIPESTATUS[1]} == 1 ]] && { UMOUNT_DEBUGFS=1; mount -t debugfs nodev /sys/kernel/debug; }

  UUID=$(cat /proc/sys/kernel/random/uuid)

  [[ -r /sys/devices/cpu/caps/pmu_name ]] && GCC_MARCH=$(cat /sys/devices/cpu/caps/pmu_name)
  [  -z "$GCC_MARCH" ] && GCC_MARCH=$(gcc -march=native -Q --help=target | grep -E '^\s*-march=' | awk -F= '{print $NF}' | awk '{$1=$1}1')
  [  -z "$GCC_MARCH" ] && GCC_MARCH=$(gcc -mtune=native -Q --help=target | grep -E '^\s*-mtune=' | awk -F= '{print $NF}' | awk '{$1=$1}1')
  [ -z "$CLOCK_SOURCE_START" ] && CLOCK_SOURCE_START="$([ -s /sys/devices/system/clocksource/clocksource0/current_clocksource ] && cat /sys/devices/system/clocksource/clocksource0/current_clocksource || echo NONE)"

  #From Power9, IBM's Power ISA supports RPT (Radix Page Table) a HPT (Hashed Page Table) MMU
  MMU=$(grep -Poh "^MMU\s*:\s*\K.+" /proc/cpuinfo)

  get_jobid_name

  {
  echo "BENCHMARK=$1"
  git_version=$(git describe --all --long 2>/dev/null) && [ -n "$git_version" ] && echo "GITVERSION=$git_version"
  git_version_long=$(git log --pretty="%h %cD %cn %s" 2>/dev/null | head -1) && [ -n "$git_version_long" ] && echo "GITVERSION_LONG=$git_version_long"
  echo "BUILD=$2"
  echo "KERNEL=${KERNEL_RPM_NAME}"
  echo "RUNNING_KERNEL=$(uname -r)"
  echo "KERNEL_FIRMWARE=$(rpm -qa|grep -E 'linux-firmware|kernel-firmware')"
  echo "MICROCODE_CTL=$(rpm -qa|grep microcode_ctl)"
  echo "GLIBC=$(rpm -qa | grep -E "glibc-[0-9]")"
  echo "MEMCPY-METHOD=$(get_memcpy_method)"
  echo "HOSTNAME=$3"
  echo "NUMASNO=$(getNumasNo)"
  echo "HT-ONOFF=$4"
  echo "RESULTTIME=$5"
  echo "PROFILE=$6"
  echo "ARGUMENTS=${ARGUMENTS[*]}"
  if [[ -n "$JOBID_NAME" ]] && [[ -n "$BEAKER_HUB_URL" ]]; then
    { echo "JOBID=$JOBID_NAME"; [[ -n "$RSTRNT_RECIPEID" ]] && echo "RECIPEID=$RSTRNT_RECIPEID"; echo "WHITEBOARD=${BEAKER_JOB_WHITEBOARD}"; echo "BEAKERHUB=${BEAKER_HUB_URL}"; }
  elif [[ -n "$BEAKER_JOB_WHITEBOARD" ]]; then
    echo "PRIVATE_WHITEBOARD=${BEAKER_JOB_WHITEBOARD}"
  fi
  echo "SOURCE_RPM_NAME=${SRC_NAME}"
  echo "HOST_TYPE=$7"
  [[ ${HOST_TYPE} == VIRTUAL ]] && [[ -n "$8" ]] && echo "VIRT_TYPE=$8"
  echo "UUID=${UUID}"
  [[ -n "$STARTDATE" ]] && echo "STARTDATE=$STARTDATE"
  echo "UPTIME_MINUS_DATE=$(get_uptime)"
  echo "PTI_ENABLED=$([ -e /sys/kernel/debug/x86/pti_enabled ] && cat /sys/kernel/debug/x86/pti_enabled || echo NONE)"
  echo "IBPB_ENABLED=$([ -e /sys/kernel/debug/x86/pti_enabled ] && cat /sys/kernel/debug/x86/ibpb_enabled || echo NONE)"
  echo "IBRS_ENABLED=$([ -e /sys/kernel/debug/x86/pti_enabled ] && cat /sys/kernel/debug/x86/ibrs_enabled || echo NONE)"
  echo "RETP_ENABLED=$([ -e /sys/kernel/debug/x86/retp_enabled ] && cat /sys/kernel/debug/x86/retp_enabled || echo NONE)"
  echo "GCC_MARCH=${GCC_MARCH:-UNKNOWN}"
  echo "MMU=${MMU:-UNKNOWN}"

  if [[ -s "parsemeta.py" ]] ; then
    if command -v python3 >/dev/null 2>&1; then
      python3 parsemeta.py
    else
      python parsemeta.py
    fi
  fi
  } >> "$CONFIG_DIR"/test_properties.txt

# shellcheck disable=SC2154 # this is intentional
  if [[ -n "$argGlabel" ]]; then
    echo "GLABEL=${argGlabel}" >> "$CONFIG_DIR"/test_properties.txt
  else
    if ! grep -q GLABEL "$CONFIG_DIR"/test_properties.txt; then
      GLABEL=$(date -u "+%Y-%m-%dT%H:%M:%S.%N" | sed 's/.\{3\}$//')
      echo "GLABEL=${GLABEL}" >> "$CONFIG_DIR"/test_properties.txt
    fi
  fi

  echo "CLOCK_SOURCE_START=${CLOCK_SOURCE_START}" >> "$CONFIG_DIR"/test_properties.txt

  [[ ${UMOUNT_DEBUGFS} == 1 ]] && umount /sys/kernel/debug
}
#}}} SaveSystemParameters

#{{{ function wait_and_get_exit_codes - sets global variable CHILD_PROCESS_ERRORS
function wait_and_get_exit_codes() {
  local children
  local code
  children=("$@")
  CHILD_PROCESS_ERRORS=0
  for job in "${children[@]}"; do
    code=0;
    wait "${job}" || code=$?
    if [[ "${code}" != "0" ]]; then
      ((CHILD_PROCESS_ERRORS+=1))
    fi
  done
  export CHILD_PROCESS_ERRORS
}
#}}}

#{{{ function rstrntUploadResult
function rstrntUploadResult()
{
  if [ -n "${RSTRNT_RECIPESETID}" ]; then
    for file in "$@"; do
      rstrnt-report-log -l "$file"
    done
  fi
}
#}}}

#{{{ function reportTestResult
# reportTestResult "$TESTNAME" "$TESTRESULT" "$OUTPUT_FILE"
# https://restraint.readthedocs.io/en/latest/commands.html
# TESTRESULT: SKIP|PASS|WARN|FAIL
function reportTestResult()
{
  local testname testresult logfile

  [ "$#" -ne 2 ] && [ "$#" -ne 3 ] && { echo "function reportTestResult - expecting 2 or 3 arguments, got $#."; printf '"%s"\n' "${@}"; echo "Returning from function."; return 1; }
  testname=$1
  testresult=$2
  [ "$#" -eq 3 ] && logfile=$3

  if [ -n "${RSTRNT_RECIPESETID}" ]; then
    if [ "$#" -eq 3 ]; then
      if [ -s "$logfile" ]; then
        rstrnt-report-result -o "$logfile" "$testname" "$testresult"
        rstrntUploadResult "$logfile"
     else
       echo 2>&1 "WARNING: function reportTestResult - file '$logfile' does not exists or is empty"
        rstrnt-report-result "$testname" "$testresult"
     fi
    else
      rstrnt-report-result "$testname" "$testresult"
    fi
  fi

  echo
  echo "=====================================RESULT for ${testname}========================================"
  echo "$testresult"
  echo "==================================================================================================="
  echo

  if [ "$#" -eq 3 ] && [ -s "$logfile" ]; then
    echo "======================================LOG for ${testname}=========================================="
    tail -v -n +1 "$logfile"
    echo "==================================================================================================="
    echo
  fi
}
#}}}

#{{{ function rsyncUploadResult $dir_name $trig_file $file1 $file2 $file3 ....
function rsyncUploadResult()
{
  local tmpd rsync_pids dir_name trig_file
  [ "$#" -lt 3 ] && { echo "function rsyncUploadResult - expecting at least 3 arguments (\$dir_name \$trig_file \$file1 \$file2 ...), got $#."; printf '"%s"\n' "${@}"; echo "Returning from function."; return 1; }
  dir_name=$1; shift
  trig_file=$1; shift

  tmpd=$(mktemp -d -t)
  mkdir -p "${tmpd}/${DISTRO_NAME}/${HOSTNAME}-${HOSTNAME_SUFFIX}/${RESULTKERNEL}/${UUID}"
  rsync_pids=""

  for rsync_server in ${RSYNC_SERVERS} ; do
    rsync_server_with_path="${rsync_server}${RSYNC_RESULTS_SPACE}"
    my_echo "rsync results to $rsync_server_with_path"
    #Create directory on server; copy results to the server
    # shellcheck disable=SC2086
    { retry ${RSYNC} "${tmpd}/" "${rsync_server_with_path}/${dir_name}" &&
      retry ${RSYNC} "${@}" "${rsync_server_with_path}/${dir_name}/${DISTRO_NAME}/${HOSTNAME}-${HOSTNAME_SUFFIX}/${RESULTKERNEL}/${UUID}" &&
      retry ${RSYNC} "$trig_file" "${rsync_server_with_path}/Preprocessing_trigger"
    } &
    rsync_pids="$rsync_pids $!"
    sleep 30
  done
  # shellcheck disable=SC2086
  wait $rsync_pids
  rm -rf "${tmpd}"
}
#}}}

#{{{ function uploadResultRsyncRstrnt $dir_name $trig_file $file1 $file2 $file3 ....
# rsyncUploadResult and rstrntUploadResult
function uploadResultRsyncRstrnt()
{
  local dir_name trig_file files
  [ "$#" -lt 3 ] && { echo "function uploadResultRsyncRstrnt - expecting at least 3 arguments (\$dir_name \$trig_file \$file1 \$file2 ...), got $#."; printf '"%s"\n' "${@}"; echo "Returning from function."; return 1; }
  dir_name=$1; shift
  trig_file=$1; shift
  files=()

  for item in "${@}"; do
    if [[ -f "${item}" ]] && [[ -s "${item}" ]]; then
      files+=("${item}")
    else
      echo "function uploadResultRsyncRstrnt - skipping '${item}' as it either do not exist or is empty"
    fi
  done

  if [ ${#files[@]} -eq 0 ]; then
    echo "function uploadResultRsyncRstrnt - no files to upload. Returning from function."
    return 1
  fi

# shellcheck disable=SC2154 # this is intentional
  [[ ${argRsync} == 1 ]] && rsyncUploadResult "$dir_name" "$trig_file" "${files[@]}"
  rstrntUploadResult "${files[@]}"
}
#}}}

#{{{ function readRsyncConfig $file
function readRsyncConfig
{
  local rsyncConfig
  [ "$#" -ne 1 ] && { echo "function readRsyncConfig - expecting 1 argument (\$rsyncConfig), got $#."; printf '"%s"\n' "${@}"; echo "Returning from function."; return 1; }
  rsyncConfig=$1
  if [[ -s "$rsyncConfig" ]]; then
    #   shellcheck disable=SC1090
    source "$rsyncConfig" || { echo "Failed to source '$rsyncConfig'"; return 1; }
    [[ -z "$RSYNC_SERVERS" ]] &&                { echo "Expecting RSYNC_SERVERS env variable to be non-empty."; return 1; }
    [[ -z "$RSYNC_KERNEL_CONFIG_SPACE" ]] &&    { echo "Expecting RSYNC_KERNEL_CONFIG_SPACE env variable to be non-empty."; return 1; }
    [[ -z "$RSYNC_RESULTS_SPACE" ]] &&          { echo "Expecting RSYNC_RESULTS_SPACE env variable to be non-empty."; return 1; }
  else
    echo "rsync_config '$rsyncConfig' file is either empty or cannot be read."
    return 1
  fi
  return 0
}
#}}}

#{{{function update_TEST_META key value
function update_TEST_META
{
  local reg

  [ "$#" -ne 2 ] && { echo "function update_TEST_META - expecting 2 arguments, got $#."; printf '"%s"\n' "${@}"; echo "Returning from function."; return 1; }
  key=$1
  value=$2
  if [[ -z "$TEST_META" ]]; then
    TEST_META="--meta $key '$value'"
  else
    reg="--meta $key '([^']*)'"
    if [[ "$TEST_META" =~ $reg ]]; then
      TEST_META=$(sed -r "s/--meta $key '[^']*'/--meta $key '$value'/" <<<"$TEST_META")
    else
      TEST_META="--meta $key '$value' $TEST_META"
    fi
  fi
}
#}}}

# {{{ function expand_range
function expand_range() {
  # input: range of numbers. Format: 1,2-5,7
  # delimiter: comma(,)
  # output: array SORTED_RANGE - 1,2,3,4,5,7
  # printf "%s\n" "${SORTED_RANGE[@]}"
  local input c array range out

  input=$1
  out=()
  export SORTED_RANGE=()
  if [[ -n "${input}" ]];then
    for c in ${input//,/ }; do
      if [[ $c =~ - ]]; then
        IFS='-' read -ra array <<< "$c"
        if (( ${#array[@]} != 2 )); then
          echo "ERROR: expand_range '$input'"
          echo        "Error when parsing '$c'. Got ${#array[@]} elements, expects 2 elements. Exiting"
          return 1
        fi
        #echo >&2 "INFO: expand_range '$input' Expanding '${c}' using 'seq ${array[0]} ${array[1]}'"
        mapfile -t range <<< $(seq "${array[0]}" "${array[1]}")
        out+=("${range[@]}")
      else
        out+=("$c")
      fi
    done
    mapfile -d '' SORTED_RANGE < <(printf '%s\0' "${out[@]}" | sort -zug)
  fi
  export SORTED_RANGE
}
# }}}

# {{{ function check_for_isolated_cores
function check_for_isolated_cores() {
  local isolated_cores_tuned isolated_cores_sys isolated_cores_tuned_a isolated_cores_sys_a
  # OUTPUT: ISOLATED_CORES_LINES
  #  0 => no isolation
  TUNED_CPU_PART_FILE=${TUNED_CPU_PART_FILE:-/etc/tuned/cpu-partitioning-variables.conf}

  if [[ -r  "$TUNED_CPU_PART_FILE" ]] ; then
    echo "check_for_isolated_cores: Found configuration file: '$TUNED_CPU_PART_FILE'"
    tail -n +1 -v "${TUNED_CPU_PART_FILE}"
    echo -e "======================================\n"
    echo "check_for_isolated_cores: Looking for '^isolated_cores=' line in '$TUNED_CPU_PART_FILE'"
    isolated_cores_tuned=$(grep "^isolated_cores=" "$TUNED_CPU_PART_FILE" | sed 's/isolated_cores=//')
    if [[ -n "${isolated_cores_tuned}" ]]; then
      echo "check_for_isolated_cores: Found cpus marked as isolated_cores in '$TUNED_CPU_PART_FILE': '$isolated_cores_tuned'"
      echo "check_for_isolated_cores: Checking /proc/cmdline"
      tail -n +1 -v /proc/cmdline
      echo -e "======================================\n"
      ISOLATED_CORES_LINES=$(grep -c "${isolated_cores_tuned}" /proc/cmdline)
      if (( ISOLATED_CORES_LINES == 0 )); then
        echo "check_for_isolated_cores: No '${isolated_cores_tuned}' in '/proc/cmdline', assuming cpu-partitioning is off"
      else
        echo "check_for_isolated_cores: Cpus '${isolated_cores_tuned}' found in '/proc/cmdline', cpu-partitioning seems to be on"
      fi
    else
      echo "check_for_isolated_cores: No isolated_cores in '$TUNED_CPU_PART_FILE'"
      ISOLATED_CORES_LINES=0
    fi
  else
    echo "check_for_isolated_cores: No '$TUNED_CPU_PART_FILE' file, assuming cpu-partitioning is off"
    ISOLATED_CORES_LINES=0
  fi

  if (( ISOLATED_CORES_LINES > 0 )); then
    # Verify with /sys/devices/system/cpu/nohz_full
    if [[ -r /sys/devices/system/cpu/nohz_full ]]; then
      echo "check_for_isolated_cores: Cross checking '/sys/devices/system/cpu/nohz_full'"
      tail -n +1 -v /sys/devices/system/cpu/nohz_full
      echo -e "======================================\n"
      isolated_cores_sys=$(cat /sys/devices/system/cpu/nohz_full)
      if [[ -n "$isolated_cores_sys" ]]; then
        if grep -E -i --silent null <<<"$isolated_cores_sys"; then
          echo "ERROR: check_for_isolated_cores: '/sys/devices/system/cpu/nohz_full' contains 'null', cpu-partitioning is off"
          echo "       This does not match what was detected with '/proc/cmdline' and '$TUNED_CPU_PART_FILE' check"
          ISOLATED_CORES_LINES=0
        else
          if expand_range "$isolated_cores_tuned"; then
            isolated_cores_tuned_a=("${SORTED_RANGE[@]}")
            echo -ne "check_for_isolated_cores: Expanded '$isolated_cores_tuned':\t"
            printf "%s " "${isolated_cores_tuned_a[@]}"; echo
          else
            echo "ERROR: check_for_isolated_cores: 'expand_range $isolated_cores_tuned' failed."
          fi

          if expand_range "$isolated_cores_sys"; then
            isolated_cores_sys_a=("${SORTED_RANGE[@]}")
            echo -ne "check_for_isolated_cores: Expanded '$isolated_cores_sys':\t"
            printf "%s " "${isolated_cores_sys_a[@]}"; echo
          else
            echo "ERROR: check_for_isolated_cores: 'expand_range $isolated_cores_sys' failed."
          fi

          if diff <(printf "%s\n" "${isolated_cores_tuned_a[@]}") <(printf "%s\n" "${isolated_cores_sys_a[@]}") >/dev/null; then
            echo "check_for_isolated_cores: Information in /proc/cmdline and /sys/devices/system/cpu/nohz_full is consistent."
            echo "  cpu-partitioning is enabled, all tests have passed."
          else
            echo "WARNING: check_for_isolated_cores: information in /proc/cmdline and /sys/devices/system/cpu/nohz_full differs!"
            echo "  Assuming that cpu-partitioning is not correctly configured"
            ISOLATED_CORES_LINES=0
          fi
        fi
      else
        echo "check_for_isolated_cores: No information in '/sys/devices/system/cpu/nohz_full'"
      fi
    else
      echo "check_for_isolated_cores: No '/sys/devices/system/cpu/nohz_full'"
    fi
  fi
  export ISOLATED_CORES_LINES
}
# }}}

# {{{ function check_grubby_for_nohz_full
# check_grubby_for_nohz_full && echo "nohz_full found. '$CHECK_GRUBBY_FOR_NOHZ_FULL_OUT'" || echo "nohz_full NOT active"
function check_grubby_for_nohz_full() {
  local grep_nohz_full
  local grep_nohz_full2
  local grep_tuned_params
  local grep_tuned_params2
  local grubby_log
  local output=1
  local nl=$'\n'
  export CHECK_GRUBBY_FOR_NOHZ_FULL_OUT=""

  grubby_log=$(grubby --info=DEFAULT)
  if [ -n "$grubby_log" ] && grep_nohz_full=$(grep nohz_full <<<"${grubby_log}"); then
    CHECK_GRUBBY_FOR_NOHZ_FULL_OUT="${grep_nohz_full}"
    output=0
  elif [ -n "$grubby_log" ] && grep_tuned_params=$(grep tuned_params <<<"${grubby_log}"); then
    #grubby --info=DEFAULT is using tuned_params => need to check /boot/grub2/grub.cfg for tuned_params definition
    grep_tuned_params2=$(find /boot -type f -exec grep -HE "^tuned_params=|^set tuned_params=" {} \;)
    if [ -n "$grep_tuned_params2" ] && grep_nohz_full2=$(grep nohz_full <<<"${grep_tuned_params2}"); then
      export CHECK_GRUBBY_FOR_NOHZ_FULL_OUT="${grep_tuned_params}${nl}${grep_nohz_full2}"
      output=0
    fi
  fi
  return "$output"
}
# }}}

# {{{ function install_st_utility
function install_st_utility() {
  if ! command -v st >/dev/null 2>&1 ; then
    if [[ -s ../Identity-setup/st.tar.bz2 ]]; then
      ${RPM_INSTALL} perl perl-ExtUtils-MakeMaker perl-Test-Simple
      if tar jxvf ../Identity-setup/st.tar.bz2
      then
        pushd st || return 1
        perl Makefile.PL
        make
        make test
        make install
        popd || return 1
      fi
    fi
  fi
  command -v st >/dev/null 2>&1 && return 0 || return 1
}
# }}}

# {{{ function load_tracepoint_kernel_module - exports TRACE_CMD_START_TIME variable.
function load_tracepoint_kernel_module() {
  local DIST_KERN start
  command -v trace-cmd >/dev/null 2>&1 || ${RPM_INSTALL} trace-cmd

  start=$SECONDS
  if trace-cmd record --quiet -e sched:sched_update_nr_running sleep 0.1; then
    export TRACE_CMD_START_TIME=$((SECONDS - start))
    echo "trace-cmd -e sched:sched_update_nr_running works!"
    echo "===============================trace-cmd report===================================================="
    trace-cmd report | tail
    echo "==================================================================================================="
    [[ -e trace.dat ]] && rm --interactive=never trace.dat
    return 0
  fi

  [[ ! -d tracepoints-helpers ]] && { command -v git >/dev/null 2>&1 || ${RPM_INSTALL} git; git clone https://github.com/auldp/tracepoints-helpers.git; }
  pushd tracepoints-helpers/sched_tp || { echo "Failed to cd tracepoints-helpers/sched_tp. Either clone from github has failed or directory structure has failed"; return 1; }
  ${RPM_INSTALL}   elfutils-libelf-devel "kernel-devel-$(uname -r)"

  kernel_sources="/usr/src/kernels/$(uname -r)"
  if ! [[ -d "$kernel_sources" ]]; then
    echo "Cannot find kernel sources at $kernel_sources needed to compile the module"
    popd || return 1
    return 1
  fi

  export DIST_KERN=DIST_KERN
  #If tracepoints-helpers.git exists, recompile the module
  [[ -r "sched_tp.ko" ]] && KERNEL_SRC="$kernel_sources" make clean
  if KERNEL_SRC="$kernel_sources" make; then
    if insmod sched_tp.ko ;then
      echo "Module sched_tp.ko was successfully loaded"
    else
      echo "Error loading sched_tp.ko module"
      popd || return 1
      return 1
    fi
  else
    echo "Error compiling kernel module"
    popd || return 1
    return 1
  fi

  popd || return 1

  start=$SECONDS
  if trace-cmd record --quiet -e sched:sched_update_nr_running sleep 0.1; then
    export TRACE_CMD_START_TIME=$((SECONDS - start))
    echo "trace-cmd record --quiet -e sched:sched_update_nr_running works!"
    echo "===============================trace-cmd report===================================================="
    trace-cmd report | tail
    echo "==================================================================================================="
    [[ -e trace.dat ]] && rm --interactive=never trace.dat
    return 0
  else
    echo "trace-cmd record --quiet -e sched:sched_update_nr_running has failed!"
    echo "Listing available 'sched:' events"
    trace-cmd list | grep sched:
    return 1
  fi
}
# }}}

# {{{ function unload_tracepoint_kernel_module
function unload_tracepoint_kernel_module() {
  if lsmod | grep -wq "^sched_tp" &> /dev/null ; then
    rmmod sched_tp
  else
    echo "Module sched_tp is not loaded"
    return 0
  fi

  if lsmod | grep -wq "^sched_tp" &> /dev/null ; then
    echo "Module sched_tp is still loaded, failed to unload it"
    return 1
  else
    echo "Module sched_tp was succesfully loaded"
    return 0
  fi
}
# }}}

#{{{ compile Stress-ng benchmark stress_version
function compile_Stress_ng_Benchmark()
{
#Compilation error: stress-prctl.c:791:66: error: ‘PIDTYPE_PID’ undeclared (first use in this function)
#is fixed by c717101e 2021-07-19 | stress-prtcl: use shim'd enum for PIDTYPE values [Colin Ian King]
  local stress_version=$1
  my_echo "Compiling benchmark ${stress_version}"
  [[ -r "${stress_version}.tar.xz" ]] || { echo "Expecting ${stress_version}.tar.xz to be readable file"; return 1; }
  tar Jxvf "${stress_version}.tar.xz"
  pushd "${stress_version}" || return 1
  make
  if [[ -x "stress-ng" ]]; then
    cp "stress-ng" ..
  else
    echo "ERROR: Compilation failed, no 'stress-ng' binary"
    return 1
  fi
  popd || return 1

  # disable systemd-coredump collection
  # See https://gitlab.com/cki-project/kernel-tests/-/blob/master/stress/stress-ng/runtest.sh
  if [[ -f "/lib/systemd/systemd" ]] ; then
    my_echo "Disabling systemd-coredump collection"
    if [[ ! -d "/etc/systemd/coredump.conf.d" ]] ; then
      mkdir /etc/systemd/coredump.conf.d
    fi
    cat >/etc/systemd/coredump.conf.d/stress-ng.conf <<EOF
[Coredump]
Storage=none
ProcessSizeMax=0
EOF
    if systemctl list-units --all | grep -qw systemd-coredump.socket ; then
      systemctl mask --now systemd-coredump.socket
    fi
  fi
}
#}}}

# {{{ function check_perf_record
# check if perf record / report works
function check_perf_record() {
  echo "$(date) Starting check_perf_record function"
  local temp_file local_exit_status

  if ! command -v perf >/dev/null 2>&1; then
    echo "check_perf_record: 'perf' utility is required but it's not installed."
    echo "Attempting to install it with '${RPM_INSTALL} perf'"
    ${RPM_INSTALL} perf
    command -v perf >/dev/null 2>&1 || { echo "Automatic installation of perf failed."; local_exit_status=1; echo "$(date) Completed check_perf_record function with return status '$local_exit_status'."; return "$local_exit_status"; }
  fi

  temp_file=$(mktemp)
  local_exit_status=0

  perf record -o "${temp_file}" sleep 1 || { echo "check_perf_record: perf record -o ${temp_file} sleep 1 has failed"; ((local_exit_status+=1)); }
  perf report -i "${temp_file}" > "${temp_file}.perf.txt" || { echo "check_perf_record: perf report -i ${temp_file} has failed"; ((local_exit_status+=1)); }
  [ -s "${temp_file}.perf.txt" ] || { echo "check_perf_record: '${temp_file}.perf.txt' does not exist or is empty"; ((local_exit_status+=1)); }
  [ -e "$temp_file" ] && rm -f "$temp_file"
  [ -e "${temp_file}.perf.txt" ]  && rm -f "${temp_file}.perf.txt"

  echo "$(date) Completed check_perf_record function with return status '$local_exit_status'"
  return "$local_exit_status"
}
# }}}
