#!/usr/bin/env bash
#vim: set expandtab cindent fdm=marker ts=2 sw=2:

# vim       :ShellCheck
# terminal  shellcheck runtest.sh

LANG=C

#{{{ trap - signal handling
# Kill the whole process group, thus killing also descendants.
# Specifying signal EXIT is useful when using set -e
# See also http://stackoverflow.com/questions/360201/how-do-i-kill-background-processes-jobs-when-my-shell-script-exits

trap_with_arg() { # from https://stackoverflow.com/a/2183063/804678
  local func="$1"; shift
  for sig in "$@"; do
# shellcheck disable=SC2064
    trap "$func $sig" "$sig"
  done
}

stop() {
  trap - SIGINT EXIT
  printf '\nFunction stop(), part of trap handling in runtest.sh: %s\n' "received $1, killing children"
  kill -s SIGINT -- -$BASHPID
}

trap_with_arg 'stop' EXIT SIGINT SIGTERM SIGHUP
#}}}

#{{{ shellcheck source=./common_library.sh
# shellcheck disable=SC1091
. "$(dirname "$(readlink -e "$0")")/common_library.sh" || { echo "Cannot load function defition from $(dirname "$(readlink -e "$0")")/common_library.sh" ; exit 1; }
#}}}

#{{{ function usage_msg
function usage_msg()
{
  printf "Usage: %s: [OPTIONS]\n\n" "$0"
  printf "  --iterations N        Number of iterations\n"
  printf "  --benchmarks <ft|mg|sp|lu|bt|is|ep|cg|ua|dc> List of benchmarks to run.\n"
  printf "                                          Multiple benchmarks can be specified using comma as delimiter\n"
  printf "  --variant             {standard|fast|full}\n"
  printf "        standard        This is the default\n"
  printf "        fast            Run smaller amount of benchmarks\n"
  printf "        full            Run all the benchmarks\n"
  printf "        cgroups         Run using different cgroups and stress-ng to create a permanent server load\n"
  printf "  --trace_mode N        Specify mode 0 - 1. Default: 0 (all tests are run without trace enabled)\n"
  printf "                            0 Run all tests without trace\n"
  printf "                            1 Run trace-cmd record --no-filter -e sched:sched_update_nr_running whenever NAS is running\n"
  printf "                              Implies --variant fast (if not explicitly set otherwise)\n"
  printf "  --keep_RawData        Do not remove RawData directory\n"
  printf "  --minimum_threads     N Minimum threads to run on. Default == number of NUMA nodes\n"
  printf "  --maximum_threads     N Maximum threads to run on. Default == number of Processing Units\n"
  printf "  --list_of_threads     List specifying numbers of threads to run test with. Use comma as delimiter.\n"
  printf "                        Example: \"4,8,16,24\"\n"
  printf "                        You can also overwrite for each number of threads a number of iterations - use\":\"\n"
  printf "                        Example: \"4:5,8:5,16,24\" - for runs with 4 and 8 threads 5 iterations will be used.\n"
  printf "                                 - for runs with 16 and 24 threads default number of iterations will be used.\n"
  printf "  --tag                 Tag the run. It will be saved in tag.txt file for later processing.\n"
  printf "  --glabel              Specify glabel or would be generated automatically\n"
  printf "  --compile_only        Only compile the benchmarks. Do not run the tests.\n"
  printf "  --cgroups             Only run cgroups test. It requires NUMA server and \"stress\" utility.\n"
  printf "                        This is using an old standalone script. Consider using --variant cgroups instead.\n"
  printf "  --test                Do not execute the benchmark itself. Instead, executes \"sleep 10\". Implies --no_rsync\n"
  printf "  --test_min            Runs small test. One iteration, maximum threads. Implies --no_rsync\n"
  printf "  --mode                default || manual. Default is \"default\". Sets manual or automatic cpu pinning \n"
  printf "  --pstree              After each subtest is started pstree command is used to display the running processes\n"
  printf "                        this is usefull to debug problems when killing various monitoring processes like\n"
  printf "                        mpstat, numastat, numatop, etc.\n"
  printf "  --hostname            Set hostname instead of getting it automatically. Used when run on the virtual machine.\n"
  printf "  --hyperthreading <ON|OFF>            Set hyperthreading status of server instead of getting it automatically.\n"
  printf "                                       Used when run on the virtual machine.\n"
  printf "  --skip_system_info    Do not run hwloc-gather-topology and copy of /sys directories \n"
  printf "  -v | --verbose        Enable verbose mode (set -x and set -v in Bash)\n"
  printf "  -h | --help           This message\n\n"
  exit 1
}

#}}} function usage_msg

#{{{ function change_test_type()
function change_test_type() {
  local RAM_SIZE
  RAM_SIZE=$(free -g | grep "Mem:" | awk '{print $2}')
  if (( RAM_SIZE < 42 )) ;then
    my_echo "Change D test size to C because of memory (RAM) constraints"
    sed -i -e 's/D/C/g' NPB_sources/config/suite_x86_64.def;
  fi
}
#}}}

#{{{ function change_mcmodel_size()
function change_mcmodel_size(){
  my_echo "Change mcmodel size"
  sed -i -e 's/medium/large/g' NPB_sources/config/make.def;
}
#}}}

# {{{ function compileSources()

function compileSources()
{
   my_echo "compileSources"
   pushd "$TESTDIR/NPB_sources" || exit
   make suite
   popd || exit
}
# }}} function compileSources()

#{{{ function untarSources
function untarSources()
{
  my_echo "untarSources"
  if [ ! -d NPB_sources ]; then
    tar -xvf "${TESTDIR}/NPB_sources.tar.xz"
  fi
}
#}}} function untarSources

# {{{ function check_runtime
# input
#    iteration_number
#    threshold in seconds
#    change (true/false)
#    TESTBINARIES
# output - global variable NEW_TESTBINARIES

function check_runtime(){
local ITER THRESHOLD CHANGE_TYPE testBinary PROGNAME LOGFILE testTime RAM_SIZE RAM_SIZE_MB MEMORY_FOR_OS_MB testTime testMemory testMemory_MB testName newTestName newTestBinary

if (( $# < 4 )); then
  echo "check_runtime - wrong number of arguments" 1>&2
  exit 1
fi

ITER=$1
shift
THRESHOLD=$1
shift
CHANGE_TYPE=$1
shift

NEW_TESTBINARIES=""

RAM_SIZE=$(free -k | grep "Mem:" | awk '{print $2}')
RAM_SIZE_MB=$(bc <<<"scale=1;$RAM_SIZE/1024.0")
MEMORY_FOR_OS_MB=$(bc <<<2*1024)  #2GB, unit is MB

for testBinary in "$@"
do
  PROGNAME=$(basename "$testBinary")
  LOGFILE="${WARMUPS_DIR}/${PROGNAME}.test_runtime_${ITER}.log"

  my_echo "$(date +"%T") Testing runtime of $testBinary with the all available CPUs"

  if OMP_DISPLAY_ENV=TRUE /usr/bin/time -v "./$testBinary" > "$LOGFILE" 2>&1 ; then
    if grep -q "Time in seconds" "$LOGFILE" ; then

      #Everything went es expected, checking the runtime
      testTime=$(grep "Time in seconds" "$LOGFILE" | sed -r "s/.* ([0-9]+)\..*/\1/")
      testMemory=$(grep "Maximum resident set size" "$LOGFILE" | awk '{print $NF}')
      testMemory_MB=$(bc <<<"scale=1;$testMemory/1024.0")
      my_echo "$testBinary has run for $testTime seconds with the all available CPUs. Maximum resident memory size was ${testMemory_MB} MB."

      if (( testTime < THRESHOLD )); then
        if [ "$CHANGE_TYPE" == "true" ]; then
# {{{ Attempt to change the variant
          # change TESTBINARIES
          testName=$(basename -s .x "$testBinary")
          newTestName=$(tr "A-Y" "B-Z" <<<"$testName")
          newTestBinary=${testBinary/$testName/$newTestName}

          # change suite.def file
          testName=${testName/./  }
          newTestName=${newTestName/./  }
          sed -i "s/$testName/$newTestName/" NPB_sources/config/suite.def
          compileSources

# Check if everything is OK
          my_echo "$(date +"%T") As the runtime of $testBinary is bellow threshold of ${THRESHOLD} seconds, we will test higher variant $newTestBinary"

          PROGNAME=$(basename "$newTestBinary")
          LOGFILE="${WARMUPS_DIR}/${PROGNAME}.test_runtime_${ITER}.log"

          if OMP_DISPLAY_ENV=TRUE /usr/bin/time -v "./$newTestBinary" > "$LOGFILE" 2>&1 ; then
            if grep -q "Time in seconds" "$LOGFILE" ; then

              #Everything went es expected, checking the runtime
              testTime=$(grep "Time in seconds" "$LOGFILE" | sed -r "s/.* ([0-9]+)\..*/\1/")
              testMemory=$(grep "Maximum resident set size" "$LOGFILE" | awk '{print $NF}')
              testMemory_MB=$(bc <<<"scale=1;$testMemory/1024.0")
              if (( testMemory < RAM_SIZE - 2 * 1024 * 1024 )); then
                my_echo "$(date +"%T") $newTestBinary has run for $testTime seconds with the all available CPUs. Maximum resident memory size was ${testMemory_MB} MB."
                my_echo "$(date +"%T") All tests are OK, changing the test type from ${testName} to ${newTestName}"
                NEW_TESTBINARIES="$NEW_TESTBINARIES ${newTestBinary}"
              else
                my_echo "$(date +"%T") WARNING: Cannot change the variant because ${newTestBinary} test has exceeded memory limit."\
                  "Maximum resident memory size was ${testMemory_MB} MB, total RAM size is $RAM_SIZE_MB MB, memory reserved for OS is $MEMORY_FOR_OS_MB MB. Check the ${LOGFILE}."
                  NEW_TESTBINARIES="$NEW_TESTBINARIES ${testBinary}"
              fi
            else
              # grep has failed for ${newTestBinary}
              my_echo "$(date +"%T") WARNING: Cannot change the variant because ${newTestBinary} test has failed (missing 'Time in seconds' line). Check the ${LOGFILE}."
              NEW_TESTBINARIES="$NEW_TESTBINARIES ${testBinary}"
            fi
          else
              # ${newTestBinary} has failed
              my_echo "$(date +"%T") WARNING: Cannot change the variant because ${newTestBinary} test has failed (nonzero return code). Check the ${LOGFILE}."
              NEW_TESTBINARIES="$NEW_TESTBINARIES ${testBinary}"
          fi
#}}}
        else
          # This is only check - do not do anything
          my_echo "$(date +"%T") WARNING: The runtime of $testBinary is bellow threshold of ${THRESHOLD} seconds, but we will not attempt to change the variant as this is only check."
          NEW_TESTBINARIES="$NEW_TESTBINARIES ./$testBinary"
        fi
      else
          # Test time is OK, no need to change the variant
          NEW_TESTBINARIES="$NEW_TESTBINARIES ./$testBinary"
      fi

    else
      # grep has failed for ${testBinary}
      my_echo "$(date +"%T") Missing 'Time in seconds' line in the '${LOGFILE}'. Excluding ${testBinary} test."
    fi
  else
   # ${testBinary} has failed
   my_echo "$(date +"%T") $testBinary has failed to run. Check the '${LOGFILE}'. Excluding ${testBinary} test."
  fi
done

}
# }}}

# {{{ Parsing parameters
argVerbose=0; argIterations=5; argIterationsSetByUser=0; argBenchmarks=""; argTraceMode=0; argMinThreads=""; argMaxThreads=""; argKeepRawData=0; argRsync="0"; argTag=""; argCompileOnly="0";argCgroups="0";argTest="0";argTestMin="0";argMode="default";argBind="0";argPstree="0";argHostname="";argHT="";argListOfThreads=""; argVariant=""; argSystemInfo=1; argGlabel=""; argRsyncConfFile="/dev/null"

if [[ "$#" -gt 0 ]]; then
  ARGUMENTS=("$@")
elif [[ -n "${TEST_PARAM}" ]]; then
  read -r -a ARGUMENTS <<<"$TEST_PARAM"
fi

#echo "${ARGUMENTS[*]}"
if (( ${#ARGUMENTS[@]} > 0 )); then
   ARGLIST=$(getopt -o 'vh' --long 'verbose,help,trace_mode:,glabel:,keep_RawData,compile_only,cgroups,iterations:,minimum_threads:,maximum_threads:,benchmarks:,variant:,tag:,test,test_min,mode:,pstree,hostname:,hyperthreading:,list_of_threads:,skip_system_info' -n "$0" -- "${ARGUMENTS[@]}") || usage_msg

   eval set -- "${ARGLIST}"
   while true
   do
# shellcheck disable=SC2034
      case $1 in
        -v|--verbose)            set -v; set -x; argVerbose=1;;
        --pstree)                argPstree=1;;
        --iterations)            shift; argIterations=$1; argIterationsSetByUser=1;;
        --minimum_threads)       shift; argMinThreads=$1;;
        --maximum_threads)       shift; argMaxThreads=$1;;
        --list_of_threads)       shift; argListOfThreads="$1";;
        --benchmarks)            shift; argBenchmarks=$1;;
        --trace_mode)            shift; argTraceMode="$1";;
        --variant)               shift; argVariant="$1";;
        --tag)                   shift; [[ -n ${argTag} ]] && argTag="${argTag} $1" || argTag=$1;;
        --glabel)                shift; argGlabel=$1;;
        --mode)                  shift; argMode=$1;;
        --keep_RawData)          argKeepRawData=1;;
        --compile_only)          argCompileOnly=1;;
        --cgroups)               argCgroups=1;;
        --test)                  argTest=1; argRsync=0;;
        --test_min)              argTestMin=1; argRsync=0; argIterations=1; argIterationsSetByUser=1;;
        --hostname)              shift; argHostname=$1;;
        --hyperthreading)        shift; argHT=$1;;
        --skip_system_info)      argSystemInfo=0;;
        -h|--help)               usage_msg;;
        --)                      break;;
        *)                       usage_msg;;
      esac
      shift
   done

   shift
   if [[ -n "$1" ]]; then
     printf "Additional arguments are not supported.\n"
     for var in "$@"; do
       echo "$var"
     done
     usage_msg
   fi
fi
#}}} Parsing parameters

#{{{ MAIN
{  # Surrounds whole script for output redirection

#{{{ Check parameters
[[ -n "$argBenchmarks" ]] && [[ -n "$argVariant" ]] && { echo >&2 "Option --benchmarks was specified together --variant ${argVariant}. Please note that option --benchmarks has precedence."; }

if ! [[ "$argTraceMode" =~ ^[0-9]+$ ]]; then
  echo "--trace_mode supports only integer arguments in range 0-1. Got '$argTraceMode'"
  usage_msg
fi

if (( argTraceMode > 1 )) || (( argTraceMode < 0 )); then
  echo "--trace_mode has to be in range 0-1"
  usage_msg
fi

if (( argTraceMode == 1 )); then
  [[ -z "$argVariant" ]] && argVariant="fast"
  my_echo "Calling load_tracepoint_kernel_module to enable sched:sched_update_nr_running trace event."
  if ! load_tracepoint_kernel_module; then
    echo "trace-cmd record --quiet --no-filter -e sched:sched_update_nr_running does not work. Listing available events"
    [[ -e trace.dat ]] && rm --interactive=never trace.dat
    trace-cmd list | grep sched:
    usage_msg
  fi
  [[ -e trace.dat ]] && rm --interactive=never trace.dat
  [[ -n "$TRACE_CMD_START_TIME" ]] && my_echo "Start time for 'trace-cmd record --quiet --no-filter -e sched:sched_update_nr_running sleep 0.1' is ${TRACE_CMD_START_TIME}s."
else
  [[ -z "$argVariant" ]] && argVariant="standard"
fi

[[ ${argVariant} == standard ]] || [[ ${argVariant} == full ]] || [[ ${argVariant} == fast ]] || [[ ${argVariant} == cgroups ]] || { echo >&2 "Unknown variant ${argVariant}"; usage_msg; }

if [[ ${argVariant} == cgroups ]]; then
  compile_Stress_ng_Benchmark stress-ng-8ddf7b75 || exit 1
  [[ "$(getNumasNo)" -gt 2 ]] && single_threaded_jobs=2 || single_threaded_jobs=1
fi

# {{{ Info about cgroups and using systemd to control cgroups
# cgcreate, cgexec and cgdelete are obsolated by systemd

# systemd-run --unit stress_1 --slice stress_1 --scope stress -c 1
# systemd-run --unit stress_2 --slice stress_2 --scope stress -c 1
# systemd-cgtop => top per groups
# systemd-cgls => list control groups

# https://access.redhat.com/documentation/en-us/red_hat_enterprise_linux/7/html/resource_management_guide/chap-using_control_groups#sec-Creating_Transient_Cgroups_with_systemd-run
# To run a command in a specified cgroup, type as root:
# systemd-run --unit=name --scope --slice=slice_name command
# The name stands for the name you want the unit to be known under. If --unit is not specified, a unit name will be generated automatically.
# It is recommended to choose a descriptive name, since it will represent the unit in the systemctl output. The name has to be unique during runtime of the unit.
# Use the optional --scope parameter to create a transient scope unit instead of service unit that is created by default.
# With the --slice option, you can make your newly created service or scope unit a member of a specified slice. Replace slice_name with the name of an existing slice
# (as shown in the output of systemctl -t slice), or create a new slice by passing a unique name. By default, services and scopes are created as members of the system.slice.
# Replace command with the command you wish to execute in the service unit. Place this command at the very end of the systemd-run syntax,
# so that the parameters of this command are not confused for parameters of systemd-run.
# }}}

if [[ -n "$argBenchmarks" ]]; then
  argBenchmarksUniq=$(echo "${argBenchmarks}" | tr ',' '\n' | sort -u | tr '\n' ' ')

  for var in ${argBenchmarksUniq}; do
     [[ ${var} == ft ]] || [[ ${var} == mg ]] || [[ ${var} == sp ]] || [[ ${var} == lu ]] || [[ ${var} == bt ]] || [[ ${var} == is ]] || [[ ${var} == ep ]] || [[ ${var} == cg ]] || [[ ${var} == ua ]] || [[ ${var} == dc ]] || { echo >&2 "Unknown benchmark ${var}"; usage_msg; }
  done
fi

if [[ ${argMode} == "manual" ]] ; then
 argBind="1"
 [[ -n ${argTag} ]] && argTag="${argTag} mode_manual_pinning" || argTag="mode_manual_pinning"
 my_echo "Running in manual CPU pinning mode"
elif [[ ${argMode} != "default" ]] ; then
 echo >&2 "Unknown mode ${argMode}"
 usage_msg
else
 my_echo "Running in default CPU pinning mode"
fi

if [ "${argHostname}xx" != "xx" ]; then
 if ! ping -c 1 -W 1 "${argHostname}"; then
    echo "WARNING: \"${argHostname}\" is not responding to ping. Please verify value of the option --hostname"
 fi
fi

if [ "${argHT}xx" != "xx" ]; then
 [[ ${argHT} == ON ]] || [[ ${argHT} == OFF ]] || { echo >&2 "Unknown --hyperthreading parameter string \"${argHT}\" - expecting either \"ON\" or \"OFF\""; usage_msg; }
 argHT="ht${argHT}"
fi

if [ "${argListOfThreads}xx" != "xx" ] && [ "${argMinThreads}xx" != "xx" ];  then
 echo >&2 "--list_of_threads is mutual exclusive with --minimum_threads"
 usage_msg
fi

if [ "${argListOfThreads}xx" != "xx" ] && [ "${argMaxThreads}xx" != "xx" ];  then
 echo >&2 "--list_of_threads is mutual exclusive with --maximum_threads"
 usage_msg
fi

if [ "${argListOfThreads}xx" != "xx" ];  then
 argListOfThreadsUniq=$(echo "${argListOfThreads}" | tr ',' '\n' | sort -nu | tr '\n' ' ')
 for var in ${argListOfThreadsUniq}; do
   for var_a in $(echo "${var}" | tr : " ") ; do
     [[ $var_a =~ ^[0-9]+$ ]] ||  { echo >&2 "--list_of_threads: \"${var_a}\" from ${var} is not a positive integer."; usage_msg; }
   done
 done
 my_echo "Following number of threads has been specified on the command line: ${argListOfThreadsUniq}"
fi

if (( argRsync == 0 )); then
  argKeepRawData=1
  my_echo "Rsync is disabled. Implies --keep_RawData"
fi
#}}} Check parameters

# {{{ Check if we are root
if [[ $EUID -ne 0 ]]; then
   echo "This script must be run as root (required by perf stat)" 1>&2
   exit 1
fi
# }}}

# {{{ Install packages
eval "$RPM_INSTALL" bc tmux rsync numactl hwloc perf numad sysstat hwloc-gui numpy scipy tuned openssl-devel numactl-devel ncurses-libs ncurses-devel gcc-gfortran python psmisc ncurses-devel numactl-devel virt-what
install_st_utility

pathadd ./
check_if_installed mpstat lstopo gcc || { echo "$(date +"%T") Missing critical tools. Aborting." | tee /dev/kmsg; exit 1; }
# }}}

# {{{ Setup environment

#sets global variables VIRT_TYPE and HOST_TYPE
get_hosttype_and_virttype

[[ -z "${argHostname}" ]] && HOSTNAME=$(hostname) || HOSTNAME=${argHostname}

#sets global variable DISTRO_NAME
get_distro_name

# }}}

#{{{ Compile benchmarks/compile only arg
TESTDIR=$(pwd)
ARCHITECTURE=$(uname --hardware-platform)
untarSources
change_test_type

pushd "$TESTDIR/NPB_sources/config" || exit
mv suite_x86_64.def suite.def
popd || exit
change_mcmodel_size

compileSources

if [[ ${argCompileOnly} == 1 ]] ; then
  reportTestResult "NAS_COMPILATION" "PASS"
  exit 0
fi

# This is a quick modification to run cgroups_test.sh from within Beaker
if [[ ${argCgroups} == 1 ]] ; then
# Compile and load kernel module which provides sched:sched_update_nr_running trace event
  if ../Scripts/tracepoint-helpers_kernel_module.sh ; then
    echo "Kernel module which provides sched:sched_update_nr_running trace event was succesfully compiled and loaded"
  else
    echo "Kernel module which provides sched:sched_update_nr_running trace event could not be compiled/loaded. Exiting."
    exit 1
  fi

  declare -a CGROUPS_ARG=()
  [[ ${argRsync} == 0 ]] && CGROUPS_ARG+=("--no_rsync")
  BEAKER_JOB_WHITEBOARD="${BEAKER_JOB_WHITEBOARD//$'\n'/}"
  BEAKER_JOB_WHITEBOARD="${BEAKER_JOB_WHITEBOARD//---/}"
  [[ "${BEAKER_JOB_WHITEBOARD}xx" != "xx" ]] && CGROUPS_ARG+=("--job_info" "${BEAKER_JOB_WHITEBOARD// /_}")
  echo "==================Calling ./cgroups_test.sh with following arguments================"
  for var in "${CGROUPS_ARG[@]}"; do echo "$var"; done
  echo "===================================================================================="
  echo
  if ./cgroups_test.sh "${CGROUPS_ARG[@]}"; then
    "./cgroups_test.sh ${CGROUPS_ARG[*]} has succesfully finished"
    exit 0
  else
    "./cgroups_test.sh ${CGROUPS_ARG[*]} has returned non-zero status"
    exit 1
  fi
fi
#}}}

#{{{ Setup benchmark
STARTDATE=$(date '+%Y-%b-%d_%Hh%Mm%Ss')

[[ ${argRsync} == 1 ]] && RSYNC="rsync -avz --no-owner --no-group --recursive --chmod=a+r,a+w,a+X,g+s" || RSYNC="true"

if [[ ${argBind} == 1 ]]; then
  pinned=".pinned"
else
  pinned=""
fi

RESULTKERNEL=${KERNEL_CONFIG_NAME}$(uname -r)${pinned}

if command -v tuned-adm >/dev/null; then
  RESULTTUNEPROFILE=$(tuned-adm active | head -1 | awk -F' ' '{print $4}')
else
  RESULTTUNEPROFILE="NOPROFILE"
fi

RESULTSTIME="Result-${STARTDATE}"
RESULT_TOP_DIR="${TESTDIR}/${RESULTKERNEL}/${RESULTTUNEPROFILE}/${RESULTSTIME}"
RESULTS_DIR="${RESULT_TOP_DIR}/RawData"
CONFIG_DIR="${RESULTS_DIR}/Config"
WARMUPS_DIR="${RESULTS_DIR}/WarmupLogs"
RESULT_DIR_MPSTAT="${RESULTS_DIR}/MPSTAT"
RESULT_DIR_MPSTAT_PER_NUMA_NODES="${RESULTS_DIR}/MPSTAT_PER_NUMA_NODES"
RESULT_DIR_PS="${RESULTS_DIR}/PS"
RESULT_DIR_PS_WITH_NODES="${RESULTS_DIR}/PS_WITH_NODES"
RESULT_DIR_TRACE="${RESULTS_DIR}/Trace_reports"
SUMMARY="${RESULTS_DIR}/Config/SUMMARY"
EXIT_STATUS="PASS"

#Set global variables RUN_NUMATOP and RUN_NUMASTAT
check_for_numatop_and_numastat

if (( RUN_NUMASTAT == 1 )); then
  RESULT_DIR_NUMASTAT="${RESULTS_DIR}/NUMASTAT"
  mkdir -p "$RESULT_DIR_NUMASTAT"
fi

if [ "${RUN_NUMATOP}" -eq 1 ]; then
  RESULT_DIR_NUMATOP="${RESULTS_DIR}/NUMATOP"
  mkdir -p "$RESULT_DIR_NUMATOP"
fi

mkdir -p "$RESULT_DIR_PS"
mkdir -p "$RESULT_DIR_PS_WITH_NODES"
mkdir -p "$RESULT_DIR_MPSTAT"
mkdir -p "$RESULT_DIR_MPSTAT_PER_NUMA_NODES"
mkdir -p "$RESULTS_DIR"
mkdir -p "$CONFIG_DIR"
mkdir -p "$WARMUPS_DIR"
(( argTraceMode == 1 )) && mkdir -p "${RESULT_DIR_TRACE}"
touch "$SUMMARY"

if [[ -f "/boot/config-$(uname -r)" ]]; then
  cp "/boot/config-$(uname -r)" "${CONFIG_DIR}/$(uname -r).config"
  for rsync_server in ${RSYNC_SERVERS}; do
    retry ${RSYNC} "/boot/config-$(uname -r)" "${rsync_server}${RSYNC_KERNEL_CONFIG_SPACE}/${RESULTKERNEL}.config"
  done
fi
reportVulnerabilities > "${CONFIG_DIR}"/vulnerabilities.txt

FULL_LOG="${CONFIG_DIR}/run.log"
# shellcheck disable=SC2012
TOP_LEVEL_LOG_FILE=$(ls -rt ./*run.log | tail -1)
#if [[ -f ${TOP_LEVEL_LOG_FILE} ]]; then
#   ln --verbose "$TOP_LEVEL_LOG_FILE" "$FULL_LOG"
#fi

touch "${CONFIG_DIR}/command_line_arguments.txt"
if (( ${#ARGUMENTS[@]} > 0 )); then
   echo "${ARGUMENTS[*]}" >>  "${CONFIG_DIR}/command_line_arguments.txt"
fi

[[ -n "$argTag" ]] && echo "$argTag" >> "$CONFIG_DIR"/tag.txt
[[ -n "$BEAKER_JOB_WHITEBOARD" ]] && store_tag "$BEAKER_JOB_WHITEBOARD" "$CONFIG_DIR"/tag.txt

[[ -z "$argHT" ]] && HTString=$(returnHTString) || HTString=${argHT}

CLOCK_SOURCE_START="$([ -s /sys/devices/system/clocksource/clocksource0/current_clocksource ] && cat /sys/devices/system/clocksource/clocksource0/current_clocksource || echo NONE)"
echo "CLOCK_SOURCE_START=${CLOCK_SOURCE_START}" | tee /dev/kmsg

output_name="NASParallel"
[[ ${argVariant} == cgroups ]] && output_name="Cgroups"
(( argTraceMode == 1 )) && output_name="${output_name}Trace"

SaveSystemParameters "${output_name}" "$DISTRO_NAME" "$HOSTNAME" "$HTString" "$RESULTSTIME" "$RESULTTUNEPROFILE" "$HOST_TYPE" "$VIRT_TYPE"

echo "storeXMLs" | tee /dev/kmsg
storeXMLs

TESTTYPES="defaultRun"
[[ ${argVariant} == cgroups ]] && TESTTYPES="mixedRun mixedRunCgroups"

if (( argTest == 1 )) ; then
  TESTBINARIES="./sleep.sh"
else
  BENCH_DIR="NPB_sources/bin"
  declare -A TEST_NAMES
  NAS_NAMES=(bt cg dc ep ft is lu mg sp ua)
  if [[ ${ARCHITECTURE} == "aarch64" ]]; then
    #Exclude bt
    NAS_NAMES=(cg dc ep ft is lu mg sp ua)
  fi
  for NAS_NAME in "${NAS_NAMES[@]}"; do
    if GREP_OUT=$(grep "^${NAS_NAME}" NPB_sources/config/suite.def) ; then
      NAS_TYPE=$(awk '{print $2}' <<< "${GREP_OUT}")
      BINARY=${BENCH_DIR}/${NAS_NAME}.${NAS_TYPE}.x
      [[ -x "${BINARY}" ]] && TEST_NAMES[${NAS_NAME}]="${BINARY}"
    fi
  done
  echo "Available tests ${TEST_NAMES[*]}"

  TESTBINARIES=""
  if [[ "${argBenchmarks}xx" == "xx" ]]; then
     case $argVariant in
       "standard") echo "standard variant of the test"; BENCHMARKS="bt ep ft is mg sp ua" ;;
       "fast")     echo "fast variant of the test";     BENCHMARKS="ep ft mg sp" ;;
       "full")     echo "full variant of the test";     BENCHMARKS="bt cg dc ep ft is lu mg sp ua" ;;
       "cgroups")  echo "cgroups variant of the test";  BENCHMARKS="lu ft mg" ;;
       *)          echo >&2 "Unknown variant ${argVariant}"; usage_msg ;;
     esac
     for var in ${BENCHMARKS}
     do
       [ ${TEST_NAMES[${var}]+xx} ] && TESTBINARIES="${TESTBINARIES} ${TEST_NAMES[$var]}" || echo "WARNING: Test name \"${var}\" does not exist"
     done
  else
     for var in ${argBenchmarksUniq}
     do
       [ ${TEST_NAMES[${var}]+xx} ] && TESTBINARIES="${TESTBINARIES} ${TEST_NAMES[$var]}" || echo "User supplied test name \"${var}\" does not exist"
     done
  fi

# {{{ Change the class of the benchmarks which run under 5 seconds
# shellcheck disable=SC2086
  check_runtime 1 5 "true" $TESTBINARIES
  TESTBINARIES=${NEW_TESTBINARIES}
#check_runtime 2 5 "false" $TESTBINARIES
#TESTBINARIES=${NEW_TESTBINARIES}

# }}}

fi

if [[ -n "${TESTBINARIES}" ]]; then
  my_echo "Following tests will be executed ${TESTBINARIES}"
else
  my_echo "No tests to execute"
  if [ -d NPB_sources/bin ]; then
    echo "ls NPB_sources/bin"
    ls NPB_sources/bin
  else
    echo "Directory with binaries does not exist"
  fi
  exit 1
fi

# }}}

# {{{ function runTest()
function runTest()
{
  local CPUS threadsArray minimal_cpu_load
  CPUS=$(getNoCpus)

   my_echo "runTest"
   if [ "${argListOfThreadsUniq}xx" == "xx" ];  then

     if [[ ${argTestMin} -eq "1" ]] ; then
      threadsArray=$(getNoCpus)
     else
      if [ "${argMinThreads}xx" == "xx" ] ; then
        argMinThreads=$(max -g "$(getNumasNo)" "$(getSocketsNo)")

# Set the minimal number of threads to create ~ 10% CPU load
# Upto 10  CPUs - 1 thread
# Upto 20  CPUs - 2 threads
# Upto 40  CPUs - 4 threads
# Upto 80  CPUs - 8 threads
# Utpo 160 CPUs - 16 threads
# Cap it to 16 threads

        minimal_cpu_load=$(ceiling_divide "$CPUS" "10")
        (( minimal_cpu_load > 16 )) && minimal_cpu_load=16
        while (( "$argMinThreads" < minimal_cpu_load )); do
          argMinThreads=$(( 2*argMinThreads ))
        done

        if [[ ${argVariant} == cgroups ]]; then
          half_cpus=$((CPUS/2))
          (( argMinThreads > half_cpus )) && argMinThreads=$half_cpus
        fi
        (( argMinThreads == 0 )) &&  argMinThreads=1
      fi

      if [  "${argMaxThreads}xx" == "xx" ] ; then
        if [[ ${argVariant} == cgroups ]]; then
          #cgroups
          if (( argTraceMode == 1 )); then
            # cgroups with trace
            #/mnt/data/www/testing/sched/results/incoming/NASParallel_cgroups_trace
            if [[ "$(getNumasNo)" -gt 2 ]]; then
              argMaxThreads=$(( $(getNoCpus) - $(getNumasNo) ))
            else
              argMaxThreads=$(( $(getNoCpus) - single_threaded_jobs - 1))
            fi
          else
            #cgroups no trace
            #/mnt/data/www/testing/sched/results/incoming/NAS_cgroups
            argMaxThreads=$(( $(getNoCpus) - single_threaded_jobs ))
          fi
        else
          #No cgroups
          if (( argTraceMode == 1 )); then
            # No cgroups and argTraceMode == 1
            #/mnt/data/www/testing/sched/results/incoming/NASParallel_trace
            if [[ "$(getNumasNo)" -gt 1 ]]; then
              argMaxThreads=$(( $(getNoCpus) - $(getNumasNo) ))
            else
              argMaxThreads=$(( $(getNoCpus) - 1))
            fi
          else
            #Default - use all CPUs (plain NAS mode) => /mnt/data/www/testing/sched/results/incoming/NASParallel
            argMaxThreads=$(getNoCpus)
          fi
        fi
        (( argMaxThreads < argMinThreads )) && argMaxThreads=${argMinThreads}
      fi #[  "${argMaxThreads}xx" == "xx" ]

      if [[ ${argVariant} == cgroups ]]; then
        my_echo "Calling generateThreadsArrayCgroups ${argMinThreads} ${argMaxThreads} ${single_threaded_jobs}"
        threadsArray=$(generateThreadsArrayCgroups "$argMinThreads" "$argMaxThreads" "${single_threaded_jobs}")
      else
        my_echo "Calling generateThreadsArray ${argMinThreads} ${argMaxThreads}"
        threadsArray=$(generateThreadsArray "$argMinThreads" "$argMaxThreads")
      fi

      #In trace mode, if number of iterations was not explicitely set by user, use 3 as default
      (( argTraceMode == 1 )) && (( argIterationsSetByUser == 0 )) && argIterations=3

     fi  #[[ ${argTestMin} -eq "1" ]]
   else
     threadsArray="${argListOfThreadsUniq}"
   fi

   my_echo "Benchmarks will be executed with the following number of threads: ${threadsArray}"

   local RUN_DIR=Run_dir
   [[ -d "${RUN_DIR}" ]] && rm -rf "${RUN_DIR}"
   mkdir "${RUN_DIR}"
   pushd "${RUN_DIR}" || exit
   RUN_DIR_FULL=$(pwd)
   XZ_THREADS=$(getNoCores)

   for testBinary in ${TESTBINARIES}
   do
      for numberThreadTemp in ${threadsArray}
      do
        if [[ ${numberThreadTemp} =~ : ]] ; then
          numberThread=$(echo "${numberThreadTemp}" | awk -F: '{print $1}')
          NOOFITERATIONS=$(echo "${numberThreadTemp}" | awk -F: '{print $2}')
          [[ $numberThread =~ ^[0-9]+$ ]]   ||  { echo >&2 "ERROR - when parsing \"${numberThreadTemp}\". Expecting format <threads>:iterations but got \"${numberThread}\". Skipping."; continue; }
          [[ $NOOFITERATIONS =~ ^[0-9]+$ ]] ||  { echo >&2 "ERROR - when parsing \"${numberThreadTemp}\". Expecting format <threads>:iterations but got \"${NOOFITERATIONS}\". Skipping."; continue; }
        else
          numberThread=${numberThreadTemp}
          NOOFITERATIONS=${argIterations}
        fi

        my_echo "${testBinary} will be executed with ${numberThread} threads and the test will be repeated ${NOOFITERATIONS} times for the ${TESTTYPES} modes."

        if [[ ${argBind} -eq 1 ]] ; then
          CPU_LIST=$("${TESTDIR}/spread.sh" "$numberThread")
          my_echo "Binding threads to the following CPUs: $CPU_LIST"
        fi


        for iterationNumber in $(seq 1 1 ${NOOFITERATIONS})
        do
          for testType in ${TESTTYPES}
          do
            runTestITSELF "$testBinary" "$iterationNumber" "$testType" "$numberThread"
            find "${RUN_DIR_FULL}" -mindepth 1 -depth -delete -print
          done
        done
      done
   done
   popd || exit
}

# }}} function runTest()

# {{{ function runTestITSELF
# parametry CestaKProgramu KteryPokusProvadime TypTestu PoleNaJakychJadrechBezime

function runTestITSELF() {
   local PROG PROGNAME LOOP_CNT TESTTYPE THREADSNO MPSTAT_LOGFILE NUMATOP_LOGFILE NUMATOP_COMMAND COMMAND logFile MAIN_PID COMMAND_LOGFILE omp_prefix ret_code testName
   PROG=$1
   PROGNAME=$(basename "$PROG")
   LOOP_CNT=$2
   TESTTYPE=$3
   THREADSNO=$4
   shift 4;
   if [[ ${argBind} -eq 1 ]] ; then
    omp_prefix=$(printf "export OMP_DISPLAY_ENV=TRUE; export OMP_NUM_THREADS=%d; export OMP_PROC_BIND=spread; export OMP_PLACES=%s;" "$THREADSNO" "${CPU_LIST}")
   else
    omp_prefix=$(printf "export OMP_DISPLAY_ENV=TRUE; export OMP_NUM_THREADS=%d;" "$THREADSNO")
   fi

   COMMAND_LOGFILE=$(printf "${RESULTS_DIR}/${PROGNAME}.${TESTTYPE}.%03dthreads.loop%02d.sh" "$THREADSNO" "$LOOP_CNT")
   touch "$COMMAND_LOGFILE"
   chmod a+x "$COMMAND_LOGFILE"
   {
   echo '#!/bin/bash -v'
#shellcheck disable=SC2016
   echo 'RESULTS_DIR=$(pwd)'
   echo 'pushd ../../../..'
   echo "THREADSNO=${THREADSNO}"
   echo "LOOP_CNT=${LOOP_CNT}"
   echo "sync; sync; echo 3 > /proc/sys/vm/drop_caches"
   } >> "$COMMAND_LOGFILE"

   sync; sync; echo 3 > /proc/sys/vm/drop_caches

# {{{ RUN mpstat
   MPSTAT_LOGFILE=$(printf "${RESULT_DIR_MPSTAT}/${PROGNAME}.${TESTTYPE}.%03dthreads.loop%02d.mpstat" "$THREADSNO" "$LOOP_CNT")
   MPSTAT_COMMAND="mpstat -P ALL 5 >& ${MPSTAT_LOGFILE} &"
#shellcheck disable=SC2016
   echo 'MPSTAT_LOGFILE=`printf "${RESULT_DIR_MPSTAT}/'"${PROGNAME}.${TESTTYPE}"'.%03dthreads.loop%02d.mpstat" $THREADSNO $LOOP_CNT`' >> "$COMMAND_LOGFILE"
#shellcheck disable=SC2016
   echo 'mpstat -P ALL 5  >& ${MPSTAT_LOGFILE} &'  >> "$COMMAND_LOGFILE"
   eval "$MPSTAT_COMMAND"
# }}}

# {{{ RUN numatop
   if (( RUN_NUMATOP == 1 )) ; then
# shellcheck disable=SC2034
     NUMATOP_LOGFILE=$(printf "${RESULT_DIR_NUMATOP}/${PROGNAME}.${TESTTYPE}.%03dthreads.loop%02d.numatop" "$THREADSNO" "$LOOP_CNT")
#shellcheck disable=SC2016
     NUMATOP_COMMAND='echo "N" | numatop -d ${NUMATOP_LOGFILE} >/dev/null 2>&1 &'
#shellcheck disable=SC2016
     echo 'NUMATOP_LOGFILE=`printf "${RESULT_DIR_NUMATOP}/'"${PROGNAME}.${TESTTYPE}"'.%03dthreads.loop%02d.numatop" $THREADSNO $LOOP_CNT`' >> "$COMMAND_LOGFILE"
#shellcheck disable=SC2016
     echo 'echo "N" | numatop -d ${NUMATOP_LOGFILE} >/dev/null 2>&1 &' >> "$COMMAND_LOGFILE"
     eval "$NUMATOP_COMMAND"
   fi
# }}}

   testName=$(printf "${PROGNAME}.${TESTTYPE}.%03dthreads.loop%02d" "$THREADSNO" "$LOOP_CNT")
   logFile="${RESULTS_DIR}/${testName}.log"
   [[ $TESTTYPE == "mixedRunCgroups" ]] && cgroup_prefix=(systemd-run --unit "NAS_${PROGNAME}" --slice "NAS_${PROGNAME}" --scope) || cgroup_prefix=()
   COMMAND="(${omp_prefix} ${COMMAND_PREFIX} ${cgroup_prefix[*]} ../${PROG}) > ${logFile} 2>&1 &"
   [[ -e trace.dat ]] && rm --interactive=never trace.dat
#shellcheck disable=SC2016
   echo 'logFile=`printf "${RESULTS_DIR}/'"${PROGNAME}"'.${TESTTYPE}.%03dthreads.loop%02d.log" $THREADSNO $LOOP_CNT`' >> "$COMMAND_LOGFILE"
   echo "${COMMAND}" >> "$COMMAND_LOGFILE"

   stress_info=""
   if [[ $TESTTYPE == "mixedRun" ]] || [[ $TESTTYPE == "mixedRunCgroups" ]]; then
     stress_info=" Running ${single_threaded_jobs} stress-ng --cpu 1 jobs in parallel."
   fi

   echo "$(date +"%T") Starting ${PROG}, iteration ${LOOP_CNT}, testType ${TESTTYPE}, number of threads ${THREADSNO}.${stress_info}" | tee /dev/kmsg
   START=$SECONDS

   stressStatus=0
   stress_log_files=()
   stress_yaml_files=()
   #Start stress-ng processes (if needed)
   if [[ $TESTTYPE == "mixedRun" ]] || [[ $TESTTYPE == "mixedRunCgroups" ]]; then
     stress_temp_pids=()
     for stress_job in $(seq ${single_threaded_jobs}); do
       [[ $TESTTYPE == "mixedRunCgroups" ]] && stress_prefix=(systemd-run --unit "stress_${stress_job}" --slice "stress_${stress_job}" --scope) || stress_prefix=()
       testNameStress="${RESULTS_DIR}/${testName}_stress-ng_${stress_job}"
       stress_log_files+=("${testNameStress}.log")
       stress_yaml_files+=("${testNameStress}.yaml")
       "${stress_prefix[@]}" ../stress-ng --cpu 1 --verbose --oomable --metrics-brief --yaml "${testNameStress}.yaml" &> "${testNameStress}.log" &
       stress_temp_pids+=($!)
       echo "${stress_prefix[*]} ../stress-ng --cpu 1 --verbose --oomable --metrics-brief --yaml ${testNameStress}.yaml &> ${testNameStress}.log &" >> "$COMMAND_LOGFILE"
     done

     #Wait for stress-ng processes to start
     startedWhile=$SECONDS
     stress_pids=()
     while (( SECONDS-startedWhile < 5 )); do
       stress_temp_pids_new=()
       for stress_pid in "${stress_temp_pids[@]}"; do
         stress_parent_pid=$(pgrep --parent "${stress_pid}")
         [ -n "${stress_parent_pid}" ] && stress_pids+=("$stress_parent_pid") || stress_temp_pids_new+=("$stress_pid")
       done
       (( ${#stress_pids[@]} == single_threaded_jobs )) && break
       stress_temp_pids=("${stress_temp_pids_new[@]}")
       sleep 0.5
     done

     if (( ${#stress_pids[@]} != single_threaded_jobs )); then
       echo "$(date +"%T") CRITICAL: stress-ng processes are not running even after $(( SECONDS-startedWhile )) seconds. Expecting $single_threaded_jobs stress-ng processes, only ${#stress_pids[@]} are running." | tee /dev/kmsg
       pstree -alp $$
       exit_message="; stress-ng process could not be started, ${PROGNAME} was not run"
       stressStatus=1
     fi
     sleep 1
   fi

   jobStatus=0
   if (( stressStatus == 0 )); then
    # Starting NAS
    eval "$COMMAND"
    MAIN_PID=$!
    echo 'MAIN_PID=$!' >> "$COMMAND_LOGFILE"

    # Wait for NAS to start
    pspid=""
    startedWhile=$SECONDS
    max_wait_time_default=10
    info_on_wait_time_message=""
    if (( argTraceMode == 1 )); then
      # On servers with many CPUs, trace-cmd record --quiet --no-filter -e sched:sched_update_nr_running <COMMAND> takes long time before starting the process.
      # load_tracepoint_kernel_module - exports TRACE_CMD_START_TIME variable
      if [[ -n "${TRACE_CMD_START_TIME}" ]]; then
        max_wait_time=$(( max_wait_time_default + TRACE_CMD_START_TIME ))
        info_on_wait_time_message="argTraceMode active: TRACE_CMD_START_TIME is ${TRACE_CMD_START_TIME}s, total wait time: ${max_wait_time}s."
      else
        max_wait_time=$(( max_wait_time_default * 2 ))
        info_on_wait_time_message="argTraceMode active: TRACE_CMD_START_TIME is not defined, total wait time set to: ${max_wait_time}s."
      fi
    else
      max_wait_time=max_wait_time_default
      info_on_wait_time_message="argTraceMode is NOT active: total wait time set to: ${max_wait_time}s."
    fi
    while ! NAS_PID=$(pgrep "$PROGNAME"); do
      if (( SECONDS-startedWhile > max_wait_time )); then
        echo "$(date +"%T") CRITICAL: Waiting for '$PROGNAME' process to start has failed - waited for $((SECONDS-startedWhile ))s without success." | tee /dev/kmsg
        echo "Increase 'max_wait_time_default' value? Current value: ${max_wait_time_default}s."
        [[ -n "$info_on_wait_time_message" ]] && echo "${info_on_wait_time_message}"
        pstree -alp $$
        jobStatus=1
        break
      fi
      ps -p $MAIN_PID >/dev/null || { echo "$(date +"%T") CRITICAL: Parent process '$MAIN_PID' no longer exists, while waiting for child processes '$PROGNAME' to start." | tee /dev/kmsg; pstree -alp $$; jobStatus=1; break; }
      sleep 0.5
    done

    if (( argPstree == 1 )) ; then
      [[ -n "${NAS_PID}" ]] && echo "NAS PID \"pgrep ${PROGNAME}\" is ${NAS_PID}" || echo "Failed to get PID of ${PROGNAME}. Probably it has failed to start"
    fi
   fi

   if (( stressStatus == 0 )) && (( jobStatus == 0 )); then
     if (( RUN_NUMASTAT == 1 )) ; then
       NUMASTAT_LOG="${RESULT_DIR_NUMASTAT}/${testName}.numastat"
       touch "$NUMASTAT_LOG"
     fi
     THREAD_LOG="${RESULT_DIR_PS}/${testName}.thread_log"
     touch "$THREAD_LOG"

     while NAS_PID=$(pgrep "$PROGNAME"); do
       all_pids=( "${NAS_PID}" )
       [[ -n "${stress_info}" ]] && all_pids+=( "${stress_pids[@]}" )
       dateString=$(date '+%Y-%b-%d_%Hh%Mm%Ss')
       psString=$(ps -L -o pid,lwp,psr,comm -p "${all_pids[@]}") || break
       if (( RUN_NUMASTAT == 1 )) ; then
         numastatString=$(numastat -p "${all_pids[@]}") || break
         { echo "$dateString"; echo "$numastatString"; } >> "$NUMASTAT_LOG"
       fi
       { echo "$dateString"; echo "$psString"; } >> "$THREAD_LOG"
       sleep 5
     done &
     pspid=$!
   fi

   if (( argPstree == 1 )) ; then
     echo "$(date +"%T") List of child processes"
     pstree -alp $BASHPID
     echo "$(date +"%T") Waiting for process ${MAIN_PID} (${COMMAND}) to finish"
   fi

   [[ -n "${MAIN_PID}" ]] && { wait ${MAIN_PID}; ret_code=$?; exit_message="Exit status of ${PROGNAME} : ${ret_code}"; }

#shellcheck disable=SC2016
   echo 'wait ${MAIN_PID}' >> "$COMMAND_LOGFILE"

   [[ -n "$pspid" ]] && wait $pspid

   (( ${#stress_pids[@]} > 0 )) && { pkill -2 stress-ng; echo 'pkill -2 stress-ng' >> "$COMMAND_LOGFILE"; }
   (( RUN_NUMATOP == 1 )) && { pkill -2 numatop; echo 'pkill -2 numatop' >> "$COMMAND_LOGFILE"; }

   pkill -2 mpstat
   { echo 'pkill -2 mpstat'; echo 'popd || exit'; } >> "$COMMAND_LOGFILE"

   echo "$(date +"%T") Completed in $(convertsecs "$((SECONDS-START))"). ${exit_message}." | tee /dev/kmsg

   # Wait for mpstat to finish
   startedWhile=$SECONDS
   sleep 0.3
   while mpstatPids=$(pgrep mpstat) || stressPids=$(pgrep stress-ng) ; do
     if (( SECONDS-startedWhile > 5 )); then
       [[ -n "$mpstatPids" ]] && echo "$(date +"%T") WARNING: mpstat (PID ${mpstatPids}) is running $((SECONDS-startedWhile ))s after sending SIGINT signal" | tee /dev/kmsg
       [[ -n "$stressPids" ]] && echo "$(date +"%T") WARNING: stress-ng (PID ${stressPids}) is running $((SECONDS-startedWhile ))s after sending SIGINT signal" | tee /dev/kmsg
       pstree -alp $$
       break
     fi
     sleep 0.5
   done
   #echo "$((SECONDS-startedWhile ))s"

   pgrep -l -P $BASHPID && { echo "$(date +"%T") Detected child processes, killing these."; pkill -P $BASHPID; }

   grepCheck=0
   if (( jobStatus == 0 )); then
    nas_value=$(grep -Poh "Mop/s total\s+=\s+\K[0-9.]+" "${logFile}")
    result_nas_key=$(printf "%s.%s_%03d" "$PROGNAME" "${TESTTYPE}" "$THREADSNO")
    if [[ -n "$nas_value" ]]; then
      echo "${testName}: $nas_value Mop/s total"
    else
      echo "File '${logFile}' - missing expected result."
      grepCheck=1
    fi

    if (( ${#stress_yaml_files[@]} > 0 )); then
      result_stress_key=$(printf "%s.%s.stress_%03d" "$PROGNAME" "${TESTTYPE}" "$THREADSNO")
      stress_value=""
      for stress_yaml_file in "${stress_yaml_files[@]}"; do
        stress_value_tmp=$(grep -Poh "bogo-ops-per-second-usr-sys-time:\s\K[0-9.]+" "${stress_yaml_file}")
        if [[ -n "$stress_value_tmp" ]]; then
          stress_value+="$stress_value_tmp "
        else
          echo "File '${stress_yaml_file}' - missing expected result."
          grepCheck=1
        fi
      done
      [[ -n "$stress_value" ]] && echo "${testName}: stress results: $stress_value bogo-ops-per-second"
    fi
   fi

   if (( argTraceMode == 1 )) && [[ -e trace.dat ]]; then
     trace-cmd report | xz --threads="${XZ_THREADS}" > "${RESULT_DIR_TRACE}/${testName}.trace.xz"
     rm --interactive=never trace.dat
   fi


   report_fail=0
   if [[ "$ret_code" -ne 0 ]] || (( jobStatus > 0 )) || (( stressStatus > 0 )) || (( grepCheck > 0 )) ; then
    {
      echo "$(date) Test ${testName} has failed."
      if (( stressStatus == 0 )); then
        [[ "$ret_code" -ne 0 ]] && echo "${COMMAND} has returned exit code ${ret_code}"
        (( jobStatus > 0 )) && { echo "Failed to start NAS benchmark."; }
      else
         echo "Failed to start stress-ng processes. Skipping $testName test."  | tee /dev/kmsg
      fi
      (( grepCheck == 1 )) && echo "Missing expected value in the output"
      tail -35 -v "$logFile" "${stress_log_files[@]}"
    } | tee --append "$SUMMARY"
    EXIT_STATUS="FAIL"
    report_fail=1
   else
     result_array[$result_nas_key]+="${nas_value} "
     [[ -n "$stress_value" ]] && result_array[$result_stress_key]+="${stress_value} "
   fi

   #Merge log files to one
   (( ${#stress_log_files[@]} > 0 )) && tail -n +1 -v "${stress_log_files[@]}" >> "${logFile}"
   (( ${#stress_yaml_files[@]} > 0 )) && tail -n +1 -v "${stress_yaml_files[@]}" >> "${logFile}"

   (( report_fail == 1 )) && reportTestResult "$testName" "FAIL" "$logFile"
}
# }}} function runTestITSELF()

#{{{function generateThreadsArray
function generateThreadsArray()
{
  local minimum_threads maximum_threads threads_array threads_new remove THREADS_LIST THREADS_ARRAY CORES CPUS I LAST DIVISIBLE_BY J DENOMINATOR THREADS_PER_CORE
  minimum_threads=$1
  maximum_threads=$2

  THREADS_LIST=$minimum_threads
  CORES=$(getNoCores)
  CPUS=$(getNoCpus)
  DIVISIBLE_BY=$(max -g "$(getNumasNo)" "$(getSocketsNo)")

# Number of threads per core is typically 2 (Hyper Threading) but can be higher number (4 on Knights Landing (Xeon Phi) CPU)
  THREADS_PER_CORE=$(getNoThreads)

# Go upto number of physical cores in the system by multiplying by 2
  for (( I=minimum_threads; I<CORES; I*=2 )); do THREADS_LIST="$THREADS_LIST $I"; done
  THREADS_LIST="$THREADS_LIST $CORES"

# If number of threads per core is >1 then test it as well
  if (( THREADS_PER_CORE > 1 )) ; then
# Divide add CORES/2 and CORES each time
# On large systems the step is CORES/4
    if (( "${CORES}" > 63 )) ; then
      DENOMINATOR=4
    else
      DENOMINATOR=2
    fi
    STEP=$(ceiling_divide "$CORES" "$DENOMINATOR")
    STEP_DIVISIBLE=$(( $(ceiling_divide "$STEP" "$DIVISIBLE_BY") * DIVISIBLE_BY ))
    LAST=$STEP
    for (( I=2*LAST; I<=CPUS; I+=CORES )); do
       THREADS_LIST="$THREADS_LIST $(seq "$LAST" ${STEP_DIVISIBLE} "${I}")"
       THREADS_LIST="$THREADS_LIST ${I}"
       LAST=${I}
     done
# On systems with large number of CPUs create one run with few CPUs idle
    if (( "${CORES}" > 20 )) ; then
      J=$(( (CORES-2)*THREADS_PER_CORE ))
#If it's NUMA system make sure it's divisible by number of NUMA nodes / CPU Sockets
      if (( DIVISIBLE_BY > 1 )) ; then
        J=$(( J/DIVISIBLE_BY ))
        J=$(( J*DIVISIBLE_BY ))
      fi
      THREADS_LIST="$THREADS_LIST $J"
    fi

# Finally add number of CPUs in the system to the list
    THREADS_LIST="$THREADS_LIST $CPUS"
  fi

#Make sure that maximum_threads is included. It can be higher than number of CPUs in the system
  THREADS_LIST="$THREADS_LIST $maximum_threads"
#  echo "Debug: THREADS_LIST (temporary): \"$THREADS_LIST\"" 1>&2

#Sort list and make it unique
# shellcheck disable=SC2086
  mapfile -t THREADS_ARRAY < <( sort_uniq $THREADS_LIST )

# Through away numbers which are higher than maximum_threads
  threads_array=()
  for I in "${THREADS_ARRAY[@]}"
  do
#echo $I
    (( I <= maximum_threads )) && (( I >= minimum_threads )) && threads_array+=("$I")
  done

  # To limit runtime, allow maximum 6 different threads configurations
  while [ ${#threads_array[@]} -gt 6 ]; do
    remove=$(( ${#threads_array[@]}-6 ))
    threads_new=()
    for I in "${!threads_array[@]}"; do
      #Remove every other element, starting from the I==1. Keep the last element.
      #Array index starts at 0.
      if (( remove>0 )) && [ $((I%2)) -eq 1 ] && [ "$I" -ne $(( ${#threads_array[@]}-1 )) ]; then
        (( remove-- ))
      else
        threads_new+=("${threads_array[$I]}")
      fi
    done
    threads_array=("${threads_new[@]}")
  done

  printf "%s " "${threads_array[@]}"
}
#}}}function generateThreadsArray

#{{{function generateThreadsArrayCgroups
function generateThreadsArrayCgroups()
{
  local minimum_threads maximum_threads single_threaded_jobs threads_array threads_array_sorted threads_new cpus numa_nodes divisor overload e
  minimum_threads=$1
  maximum_threads=$2
  single_threaded_jobs=$3

  threads_array=()
  cpus=$(getNoCpus)
  numa_nodes=$(getNumasNo)
  [ "$numa_nodes" -gt 1 ] && cpus_per_numa_nodes=$((cpus/numa_nodes)) || cpus_per_numa_nodes=1
  threads_array=("$minimum_threads" $((cpus-cpus_per_numa_nodes)) $((cpus-8*single_threaded_jobs)) $((cpus-4*single_threaded_jobs)) $((cpus-2*single_threaded_jobs)) "$maximum_threads")
  divisor=numa_nodes
  ((cpus_per_numa_nodes+4 < 2*cpus_per_numa_nodes)) && overload=4 || overload=2
  threads_array+=($((cpus-cpus_per_numa_nodes)) $((cpus-cpus_per_numa_nodes+overload)))
  while (( divisor>1 )); do
    threads_array+=($((cpus/(divisor))) $((cpus/(divisor)+overload)))
    divisor=$((divisor/2))
  done
  mapfile -t threads_array_sorted < <( sort_uniq "${threads_array[@]}" )

  threads_array=()
  for e in "${threads_array_sorted[@]}"; do
    (( e <= maximum_threads )) && (( e >= minimum_threads )) && threads_array+=("$e")
  done

  # To limit runtime, allow maximum 6 different threads configurations
  while [ ${#threads_array[@]} -gt 6 ]; do
    remove=$(( ${#threads_array[@]}-6 ))
    threads_new=()
    for I in "${!threads_array[@]}"; do
      #Remove every other element, starting from the I==1. Keep the last element.
      #Array index starts at 0.
      if (( remove>0 )) && [ $((I%2)) -eq 1 ] && [ "$I" -ne $(( ${#threads_array[@]}-1 )) ]; then
        (( remove-- ))
      else
        threads_new+=("${threads_array[$I]}")
      fi
    done
    threads_array=("${threads_new[@]}")
  done

  printf "%s " "${threads_array[@]}"
}
#}}}function generateThreadsArrayCgroups

#{{{ InitCommandPrefix
function InitCommandPrefix()
{
  local COMMAND_PREFIX="time"
  (( argTraceMode == 1 )) && COMMAND_PREFIX="${COMMAND_PREFIX} trace-cmd record --quiet --no-filter -e sched:sched_update_nr_running "
  echo "$COMMAND_PREFIX"
}
#}}}

# {{{ Run benchmark
lstopo --no-io --horiz=CORE "${CONFIG_DIR}/topo.png"
numactl --hardware > "${CONFIG_DIR}/numactl.txt"

lscpu > "${CONFIG_DIR}/lscpu.txt"
echo "$RESULTKERNEL" >> "${CONFIG_DIR}/kernel-version.txt"
rpm -qi kernel > "${CONFIG_DIR}/installed_kernels.txt"
rpm -qa > "${CONFIG_DIR}/installed_rpm_packages.txt"
rpm -qa > "${CONFIG_DIR}/installed_rpm_packages.txt"
rpm -qa --queryformat '%{installtime} (%32{installtime:date}) %-80{name} %-40{version} %-6{arch} %{DISTRIBUTION}\n' | sort -n > "${CONFIG_DIR}/installed_rpm_packages_sorted_by_date.txt"

if (( argSystemInfo == 1 )); then
  copyConfigFiles
fi

COMMAND_PREFIX=$(InitCommandPrefix)

date +%s > "$CONFIG_DIR/started"

if [[ ${argTestMin} -eq "1" ]] ; then
  my_echo "Small test. One iteration, max threads, no rsync"
fi

declare -A result_array
runTest
#Compute statistics
if [[ ${#result_array[@]} -gt 0 ]] && command -v st >/dev/null 2>&1 ; then
  mapfile -t sorted_keys < <(printf '%s\n' "${!result_array[@]}" | sort -g)
  for key in "${sorted_keys[@]}"; do
# key is constructed as "${testName}_${thread}"
    if [[ "$key" =~ ^([^_]+)_([0-9]+)$ ]]; then
      testName=${BASH_REMATCH[1]}
      thread=${BASH_REMATCH[2]}

      STATFILE="${RESULTS_DIR}/${testName}.statistics"
      if ! [ -e "${STATFILE}" ]; then
        if [[ "$testName" == *stress* ]]; then
          echo "${testName} stress-ng: bogo-ops per second statistics for different number of threads" > "$STATFILE"
        else
          echo "${testName} Mop/s total statistics for different number of threads" > "$STATFILE"
        fi
      fi
      printf "Threads:\t%s\n" "$thread" >> "${STATFILE}"
      echo "${result_array[$key]}" | tr ' ' '\n' | sed '/^$/d' | st --complete >> "${STATFILE}"
    else
      echo "WARNING: unknown key format '$key' when computing final statistics"
    fi
  done
fi

mpstat_add_NUMA_info "$RESULT_DIR_MPSTAT" "$CONFIG_DIR"/lscpu.txt "$RESULT_DIR_MPSTAT_PER_NUMA_NODES"
ps_add_NUMA_info "$RESULT_DIR_PS" "$CONFIG_DIR"/lscpu.txt "$RESULT_DIR_PS_WITH_NODES"

date +%s >> "$CONFIG_DIR/started"
mv "$CONFIG_DIR/started" "$CONFIG_DIR/done"

if [[ ${HOST_TYPE} == MACHINE ]] ; then
  HOSTNAME_SUFFIX="${HTString}"
else
  HOSTNAME_SUFFIX="virt"
fi

(( argTraceMode == 1 )) && unload_tracepoint_kernel_module

pushd "$RESULT_TOP_DIR" || exit

#Store Kernel clocksource
CLOCK_SOURCE_END="$([ -s /sys/devices/system/clocksource/clocksource0/current_clocksource ] && cat /sys/devices/system/clocksource/clocksource0/current_clocksource || echo NONE)"
echo "CLOCK_SOURCE_END=${CLOCK_SOURCE_END}" | tee --append "$CONFIG_DIR/test_properties.txt" /dev/kmsg
if [ "${CLOCK_SOURCE_START}" != "${CLOCK_SOURCE_END}" ]; then
  echo "$(date +"%T") Clock source has changed during the test. Start value: '${CLOCK_SOURCE_START}', end value: '${CLOCK_SOURCE_END}'" | tee --append "$SUMMARY"
  { nl=$'\n'; my_echo "dmesg | grep -3 'clocksource:'${nl}$(dmesg | grep -3 clocksource:)" ; } | tee --append "$SUMMARY"
  [[ "$EXIT_STATUS" == "PASS" ]] && EXIT_STATUS="WARN"
fi

#Set done filename
if [[ "$EXIT_STATUS" == "FAIL" ]]; then
  done_file="$CONFIG_DIR/done_failed"; mv "$CONFIG_DIR/done" "$done_file"
elif [[ "$EXIT_STATUS" == "WARN" ]]; then
  done_file="$CONFIG_DIR/done_warning"; mv "$CONFIG_DIR/done" "$done_file"
else
  done_file="$CONFIG_DIR/done"
fi

#Final info for test_properties.txt
COMPLETED_DATE=$(date '+%Y-%b-%d_%Hh%Mm%Ss')
RUNTIME_SECONDS=$SECONDS
RUNTIME_HUMAN_READABLE=$(convertsecs "$RUNTIME_SECONDS")
echo "COMPLETED_DATE=${COMPLETED_DATE}" | tee --append "$CONFIG_DIR/test_properties.txt" > /dev/null
echo "RUNTIME_SECONDS=${RUNTIME_SECONDS}" | tee --append "$CONFIG_DIR/test_properties.txt" "$SUMMARY" > /dev/null
echo "RUNTIME_HUMAN_READABLE=${RUNTIME_HUMAN_READABLE}" | tee --append "$CONFIG_DIR/test_properties.txt" "$SUMMARY" > /dev/null
echo "EXIT_STATUS=${EXIT_STATUS}" | tee --append "$CONFIG_DIR/test_properties.txt" "$SUMMARY" > /dev/null

#Create results.txt
cp "$CONFIG_DIR/test_properties.txt" "$CONFIG_DIR/results.txt"
echo "Summary of results" >> "$CONFIG_DIR/results.txt"
find RawData -name "*statistics" -print0 | sort -z | xargs -r0 tail -n +1 >> "$CONFIG_DIR/results.txt"

#This is needed for the standalone version of the comparator to work properly
cp "$CONFIG_DIR/test_properties.txt" .
cp "$done_file" .
cp "$CONFIG_DIR/results.txt" .
cp "$SUMMARY" .
trig_file="${UUID:=unknown}"
echo "${output_name}/${DISTRO_NAME}/${HOSTNAME}-${HOSTNAME_SUFFIX}/${RESULTKERNEL}/${UUID}" > "$trig_file"
remove_RawData=0

if tarxz RawData.tar.xz RawData/ ; then
  uploadResultRsyncRstrnt "${output_name}" "${trig_file}" "RawData.tar.xz" "$SUMMARY" "results.txt" "test_properties.txt" "$done_file"
  (( argKeepRawData == 0 )) && remove_RawData=1
else
  echo "tarxz ${RESULT_TOP_DIR}/RawData has failed!" | tee --append "$SUMMARY"
  EXIT_STATUS="FAIL"
fi

reportTestResult "SUMMARY" "$EXIT_STATUS" "$SUMMARY"
if (( remove_RawData == 1 )); then
  rm -rf RawData
  echo "Results: $(pwd)/RawData.tar.xz"
else
  echo "Results: $(pwd)/RawData"
fi
my_echo "Completed in $(convertsecs "$SECONDS")"
popd >/dev/null || exit
#}}}

#{{{ Check for any child processes
pgrep -l -P $BASHPID && { echo "Killing child processes"; pkill -P $BASHPID; }
#}}}

# Copy output of the script to file
} 2>&1 | tee "$(date '+%Y-%b-%d_%Hh%Mm%Ss')_run.log"
#}}} MAIN
